# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Click command definition for the pywbemcli command, the top level command for
the pywbemcli click tool
"""
from __future__ import absolute_import

import os
import sys
import traceback
import click
import click_repl
from copy import deepcopy
from prompt_toolkit.history import FileHistory

import pywbem
from pywbem import DEFAULT_CA_CERT_PATHS, LOGGER_SIMPLE_NAMES, \
    LOG_DESTINATIONS, DEFAULT_LOG_DESTINATION, LOG_DETAIL_LEVELS, \
    DEFAULT_LOG_DETAIL_LEVEL

from ._context_obj import ContextObj, display_click_context
from ._common import GENERAL_OPTIONS_METAVAR, TABLE_FORMATS, \
    CIM_OBJECT_OUTPUT_FORMATS
from ._pywbem_server import PywbemServer
from .config import DEFAULT_OUTPUT_FORMAT, DEFAULT_NAMESPACE, \
    PYWBEMCLI_PROMPT, PYWBEMCLI_HISTORY_FILE, DEFAULT_MAXPULLCNT, \
    DEFAULT_CONNECTION_TIMEOUT, MAX_TIMEOUT

from ._connection_repository import ConnectionRepository

__all__ = ['cli']

PYWBEM_VERSION = "pywbem, version {}".format(pywbem.__version__)

# Defaults for some options
DEFAULT_VERIFY = True  # The default is to verify
DEFAULT_TIMESTATS = False
DEFAULT_PULL_CHOICE = 'either'
USE_PULL_CHOICE = {'either': None, 'yes': True, 'no': False}

# enable -h as additional help option
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


# pylint: disable=bad-continuation
@click.group(invoke_without_command=True,
             context_settings=CONTEXT_SETTINGS,
             options_metavar=GENERAL_OPTIONS_METAVAR)
@click.option('-n', '--name', 'svr_name', type=str, metavar='NAME',
              # defaulted in code
              envvar=PywbemServer.name_envvar,
              help='Use the WBEM server defined by the WBEM connection '
                   'definition NAME. '
                   'This option is mutually exclusive with the --server and '
                   '--mock-server options, since each defines a WBEM server. '
                   'Default: EnvVar {ev}, or none.'.
                   format(ev=PywbemServer.name_envvar))
@click.option('-m', '--mock-server', type=str, multiple=True, metavar="FILE",
              # defaulted in code
              envvar=PywbemServer.mock_server_envvar,
              help='Use a mock WBEM server that is automatically created in '
                   'pywbemcli and populated with CIM objects that are defined '
                   'in the specified MOF file or Python script file. '
                   'See the pywbemcli documentation for more information. '
                   'This option may be specified multiple times, and is '
                   'mutually exclusive with the --server and --name options, '
                   'since each defines a WBEM server. '
                   'Default: EnvVar {ev}, or none.'.
                   format(ev=PywbemServer.mock_server_envvar))
@click.option('-s', '--server', type=str, metavar='URL',
              # defaulted in code
              envvar=PywbemServer.server_envvar,
              help='Use the WBEM server at the specified URL with format: '
                   '[SCHEME://]HOST[:PORT]. '
                   'SCHEME must be "https" (default) or "http". '
                   'HOST is a short or long hostname or literal IPV4/v6 '
                   'address. '
                   'PORT defaults to 5989 for https and 5988 for http. '
                   'This option is mutually exclusive with the --mock-server '
                   'and --name options, since each defines a WBEM server. '
                   'Default: EnvVar {ev}, or none.'.
                   format(ev=PywbemServer.server_envvar))
@click.option('-u', '--user', type=str, metavar='TEXT',
              # defaulted in code
              envvar=PywbemServer.user_envvar,
              help='User name for the WBEM server. '
                   'Default: EnvVar {ev}, or none.'.
                   format(ev=PywbemServer.user_envvar))
@click.option('-p', '--password', type=str, metavar='TEXT',
              # defaulted in code
              envvar=PywbemServer.password_envvar,
              help='Password for the WBEM server. '
                   'Default: EnvVar {ev}, or prompted for if --user specified.'.
                   format(ev=PywbemServer.password_envvar))
@click.option('--verify/--no-verify', 'verify', default=None,
              # defaulted in code
              envvar=PywbemServer.verify_envvar,
              help='If --verify, client verifies the X.509 server '
                   'certificate presented by the WBEM server during TLS/SSL '
                   'handshake. If --no-verify client bypasses verification. '
                   'Default: EnvVar {ev}, or "--verify".'.
                   format(ev=PywbemServer.verify_envvar))
@click.option('--ca-certs', type=str, metavar="FILE",
              default=None,  # defaulted in code
              envvar=PywbemServer.ca_certs_envvar,
              help='Path name of a file or directory containing certificates '
                   'that will be matched against the server certificate '
                   'presented by the WBEM server during TLS/SSL handshake. '
                   'Default: EnvVar {ev}, or [{dirs}].'.
                   format(ev=PywbemServer.ca_certs_envvar,
                          dirs=', '.join(DEFAULT_CA_CERT_PATHS)))
@click.option('-c', '--certfile', type=str, metavar="FILE",
              # defaulted in code
              envvar=PywbemServer.certfile_envvar,
              help='Path name of a PEM file containing a X.509 client '
                   'certificate that is used to enable TLS/SSL 2-way '
                   'authentication by presenting the certificate to the '
                   'WBEM server during TLS/SSL handshake. '
                   'Default: EnvVar {ev}, or none.'.
                   format(ev=PywbemServer.certfile_envvar))
@click.option('-k', '--keyfile', type=str, metavar='FILE',
              # defaulted in code
              envvar=PywbemServer.keyfile_envvar,
              help='Path name of a PEM file containing a X.509 private key '
                   'that belongs to the certificate in the --certfile file. '
                   'Not required if the private key is part of the '
                   '--certfile file. '
                   'Default: EnvVar {ev}, or none.'.
                   format(ev=PywbemServer.keyfile_envvar))
@click.option('-t', '--timeout', type=click.IntRange(0, MAX_TIMEOUT),
              metavar='INT',
              # defaulted in code
              envvar=PywbemServer.timeout_envvar,
              help='Client-side timeout in seconds for operations with the '
                   'WBEM server. '
                   'Default: EnvVar {ev}, or {default}.'.
                   format(ev=PywbemServer.timeout_envvar,
                          default=DEFAULT_CONNECTION_TIMEOUT))
@click.option('-U', '--use-pull', type=click.Choice(['yes', 'no', 'either']),
              # defaulted in code
              envvar=PywbemServer.use_pull_envvar,
              help='Determines whether pull operations are used for '
                   'operations with the WBEM server that return lists of '
                   'instances, as follows: '
                   '"yes" uses pull operations and fails if not supported by '
                   'the server; '
                   '"no" uses traditional operations; '
                   '"either" (default) uses pull operations if supported by '
                   'the server, and otherwise traditional operations. '
                   'Default: EnvVar {ev}, or "either".'.
                   format(ev=PywbemServer.use_pull_envvar))
@click.option('--pull-max-cnt', type=int, metavar='INT',
              # defaulted in code
              envvar=PywbemServer.pull_max_cnt_envvar,
              help='Maximum number of instances to be returned by the WBEM '
                   'server in each open or pull response, if pull operations '
                   'are used. '
                   'This is a tuning parameter that does not affect the '
                   'external behavior of the commands. '
                   'Default: EnvVar {ev}, or {default}'.
                   format(ev=PywbemServer.pull_max_cnt_envvar,
                          default=DEFAULT_MAXPULLCNT))
@click.option('-T', '--timestats', is_flag=True,
              # defaulted in code
              help='Show time statistics of WBEM server operations.')
@click.option('-d', '--default-namespace', type=str, metavar='NAMESPACE',
              default=None,
              envvar=PywbemServer.defaultnamespace_envvar,
              help='Default namespace, to be used when commands do not '
                   'specify the --namespace command option. '
                   'Default: EnvVar {ev}, or {default}.'.
                   format(ev=PywbemServer.defaultnamespace_envvar,
                          default=DEFAULT_NAMESPACE))
@click.option('-o', '--output-format', metavar='FORMAT',
              # defaulted in code
              help='Output format for the command result. '
                   'The specified format may be overriden since not all '
                   'formats apply to all result data types. '
                   'FORMAT is a table format [{tb}] or object format [{ob}]. '
                   'Default: {default}.'.
                   format(tb='|'.join(TABLE_FORMATS),
                          ob='|'.join(CIM_OBJECT_OUTPUT_FORMATS),
                          default=DEFAULT_OUTPUT_FORMAT))
@click.option('-l', '--log', type=str, metavar='COMP[=DEST[:DETAIL]],...',
              # defaulted in code
              envvar=PywbemServer.log_envvar,
              help='Enable logging of the WBEM operations, defined by a list '
                   'of log configuration strings with: '
                   'COMP: [{comp_choices}]; '
                   'DEST: [{dest_choices}], default: {dest_default}; '
                   'DETAIL: [{detail_choices}], default: {detail_default}. '
                   'Default: EnvVar {ev}, or {default}.'.
                   format(comp_choices='|'.join(LOGGER_SIMPLE_NAMES),
                          dest_choices='|'.join(LOG_DESTINATIONS),
                          dest_default=DEFAULT_LOG_DESTINATION,
                          detail_choices='|'.join(LOG_DETAIL_LEVELS),
                          detail_default=DEFAULT_LOG_DETAIL_LEVEL,
                          ev=PywbemServer.log_envvar,
                          default='all'))
@click.option('-v', '--verbose/--no-verbose',
              default=None,
              help='Display extra information about the processing.')
@click.version_option(
    message='%(prog)s, version %(version)s\n' + PYWBEM_VERSION,
    help='Show the version of this command and the pywbem package and exit.')
@click.pass_context
def cli(ctx, server, svr_name, default_namespace, user, password, timeout,
        verify, certfile, keyfile, ca_certs, output_format, use_pull,
        pull_max_cnt, mock_server, verbose=None, timestats=None, log=None):
    """
    Pywbemcli is a command line WBEM client that uses the DMTF CIM-XML protocol
    to communicate with WBEM servers. Pywbemcli can:

    \b
    * Manage the information in WBEM servers CIM objects using the
      operations defined in the DMTF specification.  It can manage CIM
      classes, CIM instances and CIM qualifier declarations in the WBEM
      Server and execute CIM methods and queries on the server.

    \b
    * Inspect WBEM server characteristics including namespaces, registered
      profiles, and other server information.

    \b
    * Capture detailed information on communication with the WBEM
      server including time statistics and logs of the operations.

    \b
    * Maintain a persistent list of named connections to WBEM servers
      and execute operations on them by name.

    Pywbemcli implements command groups and commands to execute the CIM-XML
    operations defined by the DMTF CIM Operations Over HTTP specification
    (DSP0200).

    The general options shown below can also be specified on any of the
    commands, positioned right after the 'pywbemcli' command name.

    For more detailed documentation, see:

        https://pywbemtools.readthedocs.io/en/stable/
    """
    # list of options that are not allowed in some cases:
    # i.e. When -name is used and when in interactive mode.
    conditional_options = ((default_namespace, 'default_namespace'),
                           (timeout, 'timeout'),
                           (verify, 'verify'),
                           (user, 'user'),
                           (certfile, 'certfile'),
                           (keyfile, 'keyfile'),
                           (ca_certs, 'ca_certs'),
                           (server, 'server'),
                           (mock_server, 'mock-server'))

    def create_server_instance(svr_name):
        """
        Create a PywbemServer instance from the cli arguments and return
        that object.  This method depends on using variables from the
        enclosing function. If the arguments to create a server do not exist
        no server is created.
        NOTE: svr_name passed to this function because the ClickException
        treats it as a reference before assignment if using the svr_name
        from the enclosing scopt
        """
        # test for conflicting server definitions.
        if server or mock_server:
            if svr_name:
                click.ClickException('Option conflict: --name "%s" '
                                     'conflicts with existence of --server and '
                                     '--mock-server' % svr_name)
            svr_name = 'not-saved'
            pywbem_server = PywbemServer(server,
                                         resolved_default_namespace,
                                         name=svr_name,
                                         user=user,
                                         password=password,
                                         timeout=resolved_timeout,
                                         verify=resolved_verify,
                                         certfile=certfile,
                                         keyfile=keyfile,
                                         ca_certs=resolved_ca_certs,
                                         mock_server=resolved_mock_server)
        else:  # Server and mock_server were not specified
            # if name cmd line option, get connection repo and
            # get name from the repo.
            connections = ConnectionRepository()
            local_svr_name = None
            if svr_name:
                if svr_name not in connections:
                    raise click.ClickException(
                        "Connection definition '{}' not found in "
                        "connections file '{}'".
                        format(svr_name, connections.connections_file))
                local_svr_name = svr_name
            else:  # no connection name specified
                # get the default_connection definition
                local_svr_name = connections.get_default_connection_name()
                if local_svr_name and local_svr_name not in connections:
                    connections.set_default_connection(None)
                    click.echo(
                        "Default connection definition '{}' not found in "
                        "connections file '{}'; deleted default connection "
                        "in connections file".
                        format(local_svr_name, connections.connections_file),
                        err=True)
                    raise click.Abort()
                if verbose:
                    click.echo('Current connection is "%s"' % local_svr_name)

            # Get the named connection from the repo
            if local_svr_name:
                pywbem_server = connections[local_svr_name]
                # Test for invalid options with the --name option
                for option in conditional_options:
                    if option[0]:
                        raise click.ClickException(
                            '"--%s %s" option invalid when --name exists or '
                            'default name set.' % (option[1], option[0]))

            else:
                # If no server defined, set None to allow commands that
                # do not require a server to be executed with no server
                # defined.
                pywbem_server = None
        return pywbem_server

    #
    # Process cli options to produce resolved options, i.e. the
    # options with any defaults applied for non None options.
    # Produces new variables resolved... so that later tests can confirm that
    # original variables were None or not None
    pywbem_server = None
    resolved_default_namespace = default_namespace or DEFAULT_NAMESPACE
    resolved_timestats = timestats or DEFAULT_TIMESTATS
    resolved_verify = DEFAULT_VERIFY if verify is None else verify
    resolved_ca_certs = DEFAULT_CA_CERT_PATHS if ca_certs is None else ca_certs

    if server and svr_name:
        raise click.ClickException(
            'Conflicting server definitions: name: {}, server: {}'.
            format(svr_name, server))

    if keyfile and not certfile:
        raise click.ClickException(
            'The --keyfile option "{}" is allowed only if the --certfile '
            'option is also used'.format(keyfile))
    # process mock_server option
    resolved_mock_server = []
    if mock_server:
        assert isinstance(mock_server, tuple)
        # resolve relative and absolute paths
        mock_server_path = []
        for fn in mock_server:
            if fn == os.path.basename(fn):
                mock_server_path.append(os.path.join(os.getcwd(), fn))
            else:
                mock_server_path.append(fn)

        # Test for non-existent files
        # TODO: Future: Create common method with code in build_respository
        for file_path in mock_server_path:
            ext = os.path.splitext(file_path)[1]
            if ext not in ['.py', '.mof']:
                raise click.ClickException(
                    "Mock file '{}' has invalid suffix '{}' "
                    "- must be '.py' or '.mof'".format(file_path, ext))
            if not os.path.isfile(file_path):
                raise click.ClickException(
                    "Mock file '{}' does not exist".format(file_path))
            if ext != '.py':
                resolved_mock_server.append(file_path)
                continue

            # The following allows executing selected python scripts at
            # startup but with only VERBOSE as a known global.  Inserted
            # primarily to support testing.
            with open(file_path) as fp:
                if '!PROCESS!AT!STARTUP!' in fp.readline():
                    file_source = fp.read()
                    # Only verbose is allowed here
                    globalparams = {'VERBOSE': verbose}
                    try:
                        # Using compile+exec instead of just exec allows
                        # specifying the file name, causing it to appear in
                        # any tracebacks.
                        file_code = compile(file_source, file_path, 'exec')
                        # pylint: disable=exec-used
                        exec(file_code, globalparams, None)
                    except Exception:
                        exc_type, exc_value, exc_traceback = sys.exc_info()
                        tb = traceback.format_exception(exc_type, exc_value,
                                                        exc_traceback)
                        raise click.ClickException(
                            "Mock Python process-at-startup script '{}' "
                            "failed:\n{}" %
                            (file_path, "\n".join(tb)))
                else:  # not processed during startup
                    resolved_mock_server.append(file_path)

    # Simultaneous mock_server and server fails
    if server and resolved_mock_server:
        raise click.ClickException(
            'Conflicting server definitions: server: {}, mock-server: {}'.
            format(server, ', '.join(resolved_mock_server)))

    # simultaneous mock_server and svr_name fails
    if resolved_mock_server and svr_name:
        raise click.ClickException(
            'Conflicting server definitions: mock-server: {}, name: {}'.
            format(', '.join(resolved_mock_server), svr_name))

    if use_pull:
        try:
            resolved_use_pull = USE_PULL_CHOICE[use_pull]
        except KeyError:
            raise click.ClickException(
                'Invalid choice for --use-pull %s' % use_pull)
    else:
        resolved_use_pull = DEFAULT_PULL_CHOICE

    resolved_pull_max_cnt = pull_max_cnt or DEFAULT_MAXPULLCNT

    resolved_timeout = timeout or DEFAULT_CONNECTION_TIMEOUT

    # Command mode (ctx is None). Processes command on comand line and quits
    # Apply the documented option defaults to create a pywbem_server instance
    # and a ContextObj instance
    if ctx.obj is None:  # No context. This is cmd line mode
        # Create the PywbemServer object (this contains all of the info
        # for the connection defined by the cmd line input)
        pywbem_server = create_server_instance(svr_name)

    # Interactive mode cmd line processing (ctx not Nome)
    # In interactive mode, general options specified in cmd line are used
    # to modify the pywbem_server object and the general options
    # for a single command execution.
    # This requires being able to determine for each option whether it has been
    # specified and is why general options don't define defaults in the
    # decorators that define them.
    else:  # ctx.obj exists. Processing an interactive command.
        # Apply the option defaults from the command line options
        # or from the context object.

        # If --name option, get server from connection
        if svr_name:
            connections = ConnectionRepository()
            try:
                pywbem_server = connections[svr_name]
            except KeyError:
                raise click.ClickException(
                    "Connection definition '{}' not found in "
                    "connections file '{}'".
                    format(svr_name, connections.connections_file))

        # If other parameters, modify the existing connection and reset it.
        # TODO: refactor this code to avoid deepcopy when not needed.
        else:
            if pywbem_server is None:
                # Copy to keep the original clean from any changes
                pywbem_server = deepcopy(ctx.obj.pywbem_server)
            if pywbem_server:
                modified_server = False
                if server:
                    if pywbem_server.mock_server:
                        pywbem_server.mock_server = []
                    pywbem_server.server = server
                    modified_server = True
                if mock_server:
                    if pywbem_server.server:
                        pywbem_server.server = None
                    pywbem_server.mock_server = resolved_mock_server
                    modified_server = True
                if user:
                    pywbem_server.user = user
                    modified_server = True
                if password:
                    pywbem_server.password = password
                    modified_server = True
                if verify is not None:
                    pywbem_server.verify = resolved_verify
                    modified_server = True
                if ca_certs:
                    pywbem_server.ca_certs = ca_certs
                    modified_server = True
                if certfile:
                    pywbem_server.certfile = certfile
                    modified_server = True
                if keyfile:
                    pywbem_server.keyfile = keyfile
                    modified_server = True
                if timeout:
                    pywbem_server.timeout = resolved_timeout
                if server:
                    pywbem_server.server = server
                    modified_server = True
                if default_namespace:
                    pywbem_server.default_namespace = resolved_default_namespace
                    modified_server = True
                if modified_server:
                    pywbem_server.reset()
                else:
                    pywbem_server = ctx.obj.pywbem_server
            else:
                pywbem_server = create_server_instance(svr_name)

        # The following variables are maintained only in the context_obj and
        # not attached to any particular connection. If the cli argument is
        # None, this argument was not defined as part of this interactive
        # command
        if output_format is None:
            output_format = ctx.obj.output_format
        if use_pull is None:
            resolved_use_pull = ctx.obj.use_pull
        if pull_max_cnt is None:
            resolved_pull_max_cnt = ctx.obj.pull_max_cnt
        if not timestats:  # Defaults to False, not None
            resolved_timestats = ctx.obj.timestats
        if log is None:
            log = ctx.obj.log
        if verbose is None:
            verbose = ctx.obj.verbose

    # Create a command context for each command: An interactive command has
    # its own command context as a child of the command context for the
    # command line.
    ctx.obj = ContextObj(pywbem_server, output_format,
                         resolved_use_pull,
                         resolved_pull_max_cnt,
                         resolved_timestats,
                         log, verbose)
    if verbose:
        print('CONTEXT_obj FROM generals loc 547 %r' % ctx.obj)
        print('CLICK CTX %s' % ctx)
        display_click_context(ctx, msg="After adding Context",
                              display_attrs=True)
    # Invoke command if one exists.
    if ctx.invoked_subcommand is None:
        ctx.invoke(repl)


@cli.command('help')
@click.pass_context
def repl_help(ctx):  # pylint: disable=unused-argument
    """
    Show help message for interactive mode.
    """
    click.echo("""
The following can be entered in interactive mode:

  COMMAND                     Execute pywbemcli command COMMAND.
  !SHELL-CMD                  Execute shell command SHELL-CMD.
  <CTRL-D>, :q, :quit, :exit  Exit interactive mode.
  <TAB>                       Tab completion (can be used anywhere).
  -h, --help                  Show pywbemcli general help message, including a
                              list of pywbemcli commands.
  COMMAND --help              Show help message for pywbemcli command COMMAND.
  help                        Show this help message.
  :?, :h, :help               Show help message about interactive mode.
  <UP>, <DOWN>                Scroll through pwbemcli command history.
""")


@cli.command('repl')
@click.pass_context
def repl(ctx):
    """
    Enter interactive mode (default).

    Enter the interactive mode where pywbemcli commands can be entered
    interactively. The prompt is changed to 'pywbemcli>'.

    Command history is supported. The command history is stored in a file
    ~/.pywbemcli_history.

    Pywbemcli may be terminated from this mode by entering
    <CTRL-D>, :q, :quit, :exit
    """

    history_file = PYWBEMCLI_HISTORY_FILE
    if history_file.startswith('~'):
        history_file = os.path.expanduser(history_file)

    click.echo("Enter 'help' for help, <CTRL-D> or ':q' to exit pywbemcli.")

    prompt_kwargs = {
        'message': PYWBEMCLI_PROMPT,
        'history': FileHistory(history_file),
    }
    click_repl.repl(ctx, prompt_kwargs=prompt_kwargs)



# Copyright 2017 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests the commands in thequalifier command group
"""
import os
import pytest
from .cli_test_extensions import CLITestsBase
from .common_options_help_lines import CMD_OPTION_HELP_HELP_LINE, \
    CMD_OPTION_NAMESPACE_HELP_LINE, CMD_OPTION_SUMMARY_HELP_LINE

TEST_DIR = os.path.dirname(__file__)

# A mof file that defines basic qualifier decls, classes, and instances
# but not tied to the DMTF classes.
SIMPLE_MOCK_FILE = 'simple_mock_model.mof'

QD_HELP_LINES = [
    'Usage: pywbemcli qualifier [COMMAND-OPTIONS] COMMAND [ARGS]...'
    'Command group for CIM qualifier declarations.',
    CMD_OPTION_HELP_HELP_LINE,
    'enumerate  List the qualifier declarations in a namespace.',
    'get        Get a qualifier declaration.',
]

QD_ENUMERATE_HELP_LINES = [
    'Usage: pywbemcli qualifier enumerate [COMMAND-OPTIONS]',
    'List the qualifier declarations in a namespace.',
    CMD_OPTION_NAMESPACE_HELP_LINE,
    CMD_OPTION_SUMMARY_HELP_LINE,
    CMD_OPTION_HELP_HELP_LINE,
]

QD_GET_HELP_LINES = [
    'Usage: pywbemcli qualifier get [COMMAND-OPTIONS] QUALIFIERNAME',
    "Get a qualifier declaration.",
    CMD_OPTION_NAMESPACE_HELP_LINE,
    CMD_OPTION_HELP_HELP_LINE,
]

QD_ENUM_MOCK = """Qualifier Abstract : boolean = false,
    Scope(class, association, indication),
    Flavor(EnableOverride, Restricted);

Qualifier Aggregate : boolean = false,
    Scope(reference),
    Flavor(DisableOverride, ToSubclass);

Qualifier Association : boolean = false,
    Scope(association),
    Flavor(DisableOverride, ToSubclass);

Qualifier Description : string,
    Scope(any),
    Flavor(EnableOverride, ToSubclass, Translatable);

Qualifier In : boolean = true,
    Scope(parameter),
    Flavor(DisableOverride, ToSubclass);

Qualifier Indication : boolean = false,
    Scope(class, indication),
    Flavor(DisableOverride, ToSubclass);

Qualifier Key : boolean = false,
    Scope(property, reference),
    Flavor(DisableOverride, ToSubclass);

Qualifier Out : boolean = false,
    Scope(parameter),
    Flavor(DisableOverride, ToSubclass);

Qualifier Override : string,
    Scope(property, reference, method),
    Flavor(EnableOverride, Restricted);

"""

QD_GET_MOCK = """Qualifier Description : string,
    Scope(any),
    Flavor(EnableOverride, ToSubclass, Translatable);

"""
# pylint: disable=line-too-long
QD_GET_MOCK_XML = """<QUALIFIER.DECLARATION ISARRAY="false" NAME="Description" OVERRIDABLE="true" TOSUBCLASS="true" TRANSLATABLE="true" TYPE="string">
    <SCOPE ASSOCIATION="true" CLASS="true" INDICATION="true" METHOD="true" PARAMETER="true" PROPERTY="true" REFERENCE="true"/>
</QUALIFIER.DECLARATION>

"""  # noqa: E501

QD_TBL_OUT = """Qualifier Declarations
+-------------+---------+---------+---------+-------------+-----------------+
| Name        | Type    | Value   | Array   | Scopes      | Flavors         |
+=============+=========+=========+=========+=============+=================+
| Abstract    | boolean | False   | False   | CLASS       | EnableOverride  |
|             |         |         |         | ASSOCIATION | Restricted      |
|             |         |         |         | INDICATION  |                 |
+-------------+---------+---------+---------+-------------+-----------------+
| Aggregate   | boolean | False   | False   | REFERENCE   | DisableOverride |
|             |         |         |         |             | ToSubclass      |
+-------------+---------+---------+---------+-------------+-----------------+
| Association | boolean | False   | False   | ASSOCIATION | DisableOverride |
|             |         |         |         |             | ToSubclass      |
+-------------+---------+---------+---------+-------------+-----------------+
| Description | string  |         | False   | ANY         | EnableOverride  |
|             |         |         |         |             | ToSubclass      |
|             |         |         |         |             | Translatable    |
+-------------+---------+---------+---------+-------------+-----------------+
| In          | boolean | True    | False   | PARAMETER   | DisableOverride |
|             |         |         |         |             | ToSubclass      |
+-------------+---------+---------+---------+-------------+-----------------+
| Indication  | boolean | False   | False   | CLASS       | DisableOverride |
|             |         |         |         | INDICATION  | ToSubclass      |
+-------------+---------+---------+---------+-------------+-----------------+
| Key         | boolean | False   | False   | PROPERTY    | DisableOverride |
|             |         |         |         | REFERENCE   | ToSubclass      |
+-------------+---------+---------+---------+-------------+-----------------+
| Out         | boolean | False   | False   | PARAMETER   | DisableOverride |
|             |         |         |         |             | ToSubclass      |
+-------------+---------+---------+---------+-------------+-----------------+
| Override    | string  |         | False   | PROPERTY    | EnableOverride  |
|             |         |         |         | REFERENCE   | Restricted      |
|             |         |         |         | METHOD      |                 |
+-------------+---------+---------+---------+-------------+-----------------+
"""

QD_TBL_GET_OUT = """Qualifier Declarations
+----------+---------+---------+---------+-------------+----------------+
| Name     | Type    | Value   | Array   | Scopes      | Flavors        |
+==========+=========+=========+=========+=============+================+
| Abstract | boolean | False   | False   | CLASS       | EnableOverride |
|          |         |         |         | ASSOCIATION | Restricted     |
|          |         |         |         | INDICATION  |                |
+----------+---------+---------+---------+-------------+----------------+
"""

# TODO: Add tests for xml, repr, txt formats.

# The following variables are used to control tests executed during
# development of tests
OK = True      # set to OK for tests passed. Set OK = False to execute one test
RUN = True     # set RUN condition in test being run
FAIL = False   # flag any tests that fail

TEST_CASES = [
    # desc - Description of test
    # inputs - String, or list of args or dict of 'env', 'args', 'general',
    #          and 'stdin'. See See CLITestsBase.command_test()  for
    #          detailed documentation
    # exp_response - Dictionary of expected responses (stdout, stderr, rc) and
    #                test definition (test: <testname>).
    #                See CLITestsBase.command_test() for detailed documentation.
    # mock - None or name of files (mof or .py),
    # condition - If True, the test is executed,  Otherwise it is skipped.

    ['Verify qualifier command --help response',
     '--help',
     {'stdout': QD_HELP_LINES,
      'test': 'innows'},
     None, OK],

    ['Verify qualifier command -h response',
     '-h',
     {'stdout': QD_HELP_LINES,
      'test': 'innows'},
     None, OK],

    ['Verify qualifier command enumerate --help response',
     ['enumerate', '--help'],
     {'stdout': QD_ENUMERATE_HELP_LINES,
      'test': 'innows'},
     None, OK],

    ['Verify qualifier command enumerate -h response.',
     ['enumerate', '-h'],
     {'stdout': QD_ENUMERATE_HELP_LINES,
      'test': 'innows'},
     None, OK],

    ['Verify qualifier command get --help response.',
     ['get', '--help'],
     {'stdout': QD_GET_HELP_LINES,
      'test': 'innows'},
     None, OK],

    ['Verify qualifier command get -h response.',
     ['get', '-h'],
     {'stdout': QD_GET_HELP_LINES,
      'test': 'innows'},
     None, OK],

    ['Verify qualifier command enumerate returns qual decls.',
     ['enumerate'],
     {'stdout': QD_ENUM_MOCK,
      'test': 'lines'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify qualifier command enumerate with namespace returns qual decls.',
     ['enumerate', '--namespace', 'root/cimv2'],
     {'stdout': QD_ENUM_MOCK,
      'test': 'lines'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify qualifier command enumerate summary returns qual decls.',
     ['enumerate', '--summary'],
     {'stdout': ['9', 'CIMQualifierDeclaration'],
      'test': 'in'},
     SIMPLE_MOCK_FILE, OK],


    ['Verify qualifier command enumerate summary returns qual decls table',
     {'args': ['enumerate', '--summary'],
      'general': ['--output-format', 'table']},
     {'stdout': ["""Summary of CIMQualifierDeclaration returned
+---------+-------------------------+
|   Count | CIM Type                |
|---------+-------------------------|
|       9 | CIMQualifierDeclaration |
+---------+-------------------------+
"""],
      'test': 'linesnows'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify qualifier command get  Description',
     ['get', 'Description'],
     {'stdout': QD_GET_MOCK,
      'test': 'lines'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify qualifier command get invalid qual decl name .',
     ['get', 'NoSuchQualDecl'],
     {'stderr': ["Error: CIMError: 6"],
      'rc': 1,
      'test': 'in'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify qualifier command get  Description outputformat xml',
     {'args': ['get', 'Description'],
      'general': ['--output-format', 'xml']},
     {'stdout': QD_GET_MOCK_XML,
      'test': 'lines'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify qualifier command -o grid enumerate produces table out',
     {'args': ['enumerate'],
      'general': ['-o', 'grid']},
     {'stdout': QD_TBL_OUT,
      'rc': 0,
      'test': 'lines'},
     SIMPLE_MOCK_FILE, OK],


    ['Verify qualifier command -o grid get Abstract table out',
     {'args': ['get', 'abstract'],
      'general': ['-o', 'grid']},
     {'stdout': QD_TBL_GET_OUT,
      'rc': 0,
      'test': 'lines'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify qualifier command enumerate invalid namespace Fails',
     ['enumerate', '--namespace', 'root/blah'],
     {'stderr': ["Error: CIMError: 3", "CIM_ERR_INVALID_NAMESPACE"],
      'rc': 1,
      'test': 'in'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify qualifier command --timestats gets stats output. Cannot test'
     'with lines because execution time is variable.',
     {'args': ['get', 'IN'],
      'general': ['--timestats']},
     {'stdout': ['Qualifier In : boolean = true,',
                 'Scope(parameter),',
                 'Count    Exc    Time    ReqLen    ReplyLen  Operation',
                 '      1      0',
                 '0           0  GetQualifier'],
      'rc': 0,
      'test': 'in'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify qualifier command -o repr get Description produces repr out',
     {'args': ['get', 'Description'],
      'general': ['-o', 'repr']},
     {'stdout': "CIMQualifierDeclaration(name='Description', value=None, "
                "type='string', is_array=False, array_size=None, "
                "scopes=NocaseDict({'CLASS': False, 'ASSOCIATION': False, "
                "'INDICATION': False, 'PROPERTY': False, 'REFERENCE': False, "
                "'METHOD': False, 'PARAMETER': False, 'ANY': True}), "
                "tosubclass=True, overridable=True, translatable=True, "
                "toinstance=None)",
      'rc': 0,
      'test': 'lines'},
     SIMPLE_MOCK_FILE, OK],
]


class TestcmdQualifiers(CLITestsBase):
    """
    Test all of the qualifiers command variations.
    """
    command_group = 'qualifier'

    @pytest.mark.parametrize(
        "desc, inputs, exp_response, mock, condition", TEST_CASES)
    def test_qualdecl(self, desc, inputs, exp_response, mock, condition):
        """
        Common test method for those commands and options in the
        qualifier command group that can be tested.  This includes:

          * Subcommands like help that do not require access to a server

          * Subcommands that can be tested with a single execution of a
            pywbemcli command.
        """
        self.command_test(desc, self.command_group, inputs, exp_response,
                          mock, condition)



# Copyright 2018 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests the class command
"""
import os
import pytest
from .cli_test_extensions import CLITestsBase
from .common_options_help_lines import CMD_OPTION_NAMES_ONLY_HELP_LINE, \
    CMD_OPTION_HELP_HELP_LINE, CMD_OPTION_SUMMARY_HELP_LINE, \
    CMD_OPTION_NAMESPACE_HELP_LINE, CMD_OPTION_PROPERTYLIST_HELP_LINE, \
    CMD_OPTION_INCLUDE_CLASSORIGIN_HELP_LINE, \
    CMD_OPTION_LOCAL_ONLY_CLASS_HELP_LINE, CMD_OPTION_NO_QUALIFIERS_HELP_LINE, \
    CMD_OPTION_MULTIPLE_NAMESPACE_HELP_LINE

from .utils import execute_pywbemcli, assert_rc

TEST_DIR = os.path.dirname(__file__)

# A mof file that defines basic qualifier decls, classes, and instances
# but not tied to the DMTF classes.
SIMPLE_MOCK_FILE = 'simple_mock_model.mof'
SIMPLE_MOCK_FILE_EXT = 'simple_mock_model_ext.mof'
INVOKE_METHOD_MOCK_FILE = 'simple_mock_invokemethod.py'
SIMPLE_ASSOC_MOCK_FILE = 'simple_assoc_mock_model.mof'


#
# The following list define the help for each command in terms of particular
# parts of lines that are to be tested.
# For each test, try to include:
# 1. The usage line and in particular the argument component
# 2. The single
# 2. The last line CMD_OPTION_HELP_HELP_LINE
#
CLASS_HELP_LINES = [
    'Usage: pywbemcli class [COMMAND-OPTIONS] COMMAND [ARGS]...',
    'Command group for CIM classes.',
    CMD_OPTION_HELP_HELP_LINE,
    'associators   List the classes associated with a class.',
    'delete        Delete a class.',
    'enumerate     List top classes or subclasses of a class in a namespace.',
    'find          List the classes with matching class names on the server.',
    'get           Get a class.',
    'invokemethod  Invoke a method on a class.',
    'references    List the classes referencing a class.',
    'tree          Show the subclass or superclass hierarchy for a class.',
]

CLASS_ASSOCIATORS_HELP_LINES = [
    'Usage: pywbemcli class associators [COMMAND-OPTIONS] CLASSNAME',
    'List the classes associated with a class.',
    '--ac, --assoc-class CLASSNAME Filter the result set by association class',
    '--rc, --result-class CLASSNAME Filter the result set by result class',
    '-r, --role PROPERTYNAME Filter the result set by source end role',
    '--rr, --result-role PROPERTYNAME Filter the result set by far end role',
    CMD_OPTION_NO_QUALIFIERS_HELP_LINE,
    CMD_OPTION_INCLUDE_CLASSORIGIN_HELP_LINE,
    CMD_OPTION_PROPERTYLIST_HELP_LINE,
    CMD_OPTION_NAMES_ONLY_HELP_LINE,
    CMD_OPTION_NAMESPACE_HELP_LINE,
    CMD_OPTION_SUMMARY_HELP_LINE,
    CMD_OPTION_HELP_HELP_LINE,
]

CLASS_DELETE_HELP_LINES = [
    'Usage: pywbemcli class delete [COMMAND-OPTIONS] CLASSNAME',
    'Delete a class.',
    '-f, --force Delete any instances of the class as well.',
    CMD_OPTION_NAMESPACE_HELP_LINE,
    CMD_OPTION_HELP_HELP_LINE,
]

CLASS_ENUMERATE_HELP_LINES = [
    'Usage: pywbemcli class enumerate [COMMAND-OPTIONS] CLASSNAME',
    'List top classes or subclasses of a class in a namespace.',
    '--di, --deep-inheritance Include the complete subclass hierarchy',
    CMD_OPTION_LOCAL_ONLY_CLASS_HELP_LINE,
    CMD_OPTION_NO_QUALIFIERS_HELP_LINE,
    CMD_OPTION_INCLUDE_CLASSORIGIN_HELP_LINE,
    CMD_OPTION_NAMES_ONLY_HELP_LINE,
    CMD_OPTION_NAMESPACE_HELP_LINE,
    CMD_OPTION_SUMMARY_HELP_LINE,
    CMD_OPTION_HELP_HELP_LINE,
]

CLASS_FIND_HELP_LINES = [
    'Usage: pywbemcli class find [COMMAND-OPTIONS] CLASSNAME-GLOB',
    'List the classes with matching class names on the server.',
    '-s, --sort                 Sort by namespace. Default is to sort by',
    CMD_OPTION_MULTIPLE_NAMESPACE_HELP_LINE,
    CMD_OPTION_HELP_HELP_LINE,
]

CLASS_GET_HELP_LINES = [
    'Usage: pywbemcli class get [COMMAND-OPTIONS] CLASSNAME',
    'Get a class.',
    CMD_OPTION_LOCAL_ONLY_CLASS_HELP_LINE,
    CMD_OPTION_NO_QUALIFIERS_HELP_LINE,
    CMD_OPTION_INCLUDE_CLASSORIGIN_HELP_LINE,
    CMD_OPTION_PROPERTYLIST_HELP_LINE,
    CMD_OPTION_NAMESPACE_HELP_LINE,
    CMD_OPTION_HELP_HELP_LINE,
]

CLASS_INVOKEMETHOD_HELP_LINES = [
    'Usage: pywbemcli class invokemethod [COMMAND-OPTIONS] CLASSNAME '
    'METHODNAME',
    'Invoke a method on a class.',
    '-p, --parameter PARAMETERNAME=VALUE Specify a method input parameter',
    CMD_OPTION_NAMESPACE_HELP_LINE,
    CMD_OPTION_HELP_HELP_LINE,
]

CLASS_REFERENCES_HELP_LINES = [
    'Usage: pywbemcli class references [COMMAND-OPTIONS] CLASSNAME',
    'List the classes referencing a class.',
    '--rc, --result-class CLASSNAME Filter the result set by result class',
    '-r, --role PROPERTYNAME Filter the result set by source end role',
    CMD_OPTION_NO_QUALIFIERS_HELP_LINE,
    CMD_OPTION_INCLUDE_CLASSORIGIN_HELP_LINE,
    CMD_OPTION_PROPERTYLIST_HELP_LINE,
    CMD_OPTION_NAMES_ONLY_HELP_LINE,
    CMD_OPTION_NAMESPACE_HELP_LINE,
    CMD_OPTION_SUMMARY_HELP_LINE,
    CMD_OPTION_HELP_HELP_LINE,
]

CLASS_TREE_HELP_LINES = [
    'Usage: pywbemcli class tree [COMMAND-OPTIONS] CLASSNAME',
    'Show the subclass or superclass hierarchy for a class.',
    '-s, --superclasses Show the superclass hierarchy.',
    CMD_OPTION_NAMESPACE_HELP_LINE,
    CMD_OPTION_HELP_HELP_LINE,
]

CIMFOO_SUB_SUB = """
   [Description ( "Subclass of CIM_Foo_sub" )]
class CIM_Foo_sub_sub : CIM_Foo_sub {

   string cimfoo_sub_sub;

   string cimfoo_sub;

      [Key ( true ),
       Description ( "This is key property." )]
   string InstanceID;

      [Description ( "This is Uint32 property." )]
   uint32 IntegerProp;

      [Description ( "Sample method with input and output parameters" )]
   uint32 Method1(
         [IN ( false ),
          OUT ( true ),
          Description ( "Response param 2" )]
      string OutputParam2);

      [Description ( "Method with in and out parameters" )]
   uint32 Fuzzy(
         [IN ( true ),
          OUT ( true ),
          Description ( "Define data to be returned in output parameter" )]
      string TestInOutParameter,
         [IN ( true ),
          OUT ( true ),
          Description ( "Test of ref in/out parameter" )]
      CIM_Foo REF TestRef,
         [IN ( false ),
          OUT ( true ),
          Description ( "Rtns method name if exists on input" )]
      string OutputParam,
         [IN ( true ),
          Description ( "Defines return value if provided." )]
      uint32 OutputRtnValue);

      [Description ( "Method with no Parameters" )]
   uint32 DeleteNothing();

};

"""

CIMFOO_SUB_SUB_NO_QUALS = """
class CIM_Foo_sub_sub : CIM_Foo_sub {

   string cimfoo_sub_sub;

   string cimfoo_sub;

   string InstanceID;

   uint32 IntegerProp;

   uint32 Method1(
      string OutputParam2);

   uint32 Fuzzy(
      string TestInOutParameter,
      CIM_Foo REF TestRef,
      string OutputParam,
      uint32 OutputRtnValue);

   uint32 DeleteNothing();

};
"""
REFERENCES_CLASS_RTN = [
    '//FakedUrl/root/cimv2:TST_Lineage',
    'class TST_Lineage {',
    '',
    '   string InstanceID;',
    '',
    '   TST_Person REF parent;',
    '',
    '   TST_Person REF child;',
    '',
    '};',
    '',
    '//FakedUrl/root/cimv2:TST_MemberOfFamilyCollection',
    'class TST_MemberOfFamilyCollection {',
    '',
    '   TST_Person REF family;',
    '',
    '   TST_Person REF member;',
    '',
    '};',
    '', ]

REFERENCES_CLASS_RTN2 = [
    '//FakedUrl/root/cimv2:TST_MemberOfFamilyCollection',
    'class TST_MemberOfFamilyCollection {',
    '',
    '   TST_Person REF family;',
    '',
    '   TST_Person REF member;',
    '',
    '};',
    '',
    '']

REFERENCES_CLASS_RTN_QUALS1 = """
//FakedUrl/root/cimv2:TST_Lineage
   [Association ( true ),
    Description (
       " Lineage defines the relationship between parents and children." )]
class TST_Lineage {

      [key ( true )]
   string InstanceID;

   TST_Person REF parent;

   TST_Person REF child;

};

//FakedUrl/root/cimv2:TST_MemberOfFamilyCollection
   [Association ( true ),
    Description ( " Family gathers person to family." )]
class TST_MemberOfFamilyCollection {

      [key ( true )]
   TST_Person REF family;

      [key ( true )]
   TST_Person REF member;

};
"""

REFERENCES_CLASS_RTN_QUALS2 = """
//FakedUrl/root/cimv2:TST_MemberOfFamilyCollection
   [Association ( true ),
    Description ( " Family gathers person to family." )]
class TST_MemberOfFamilyCollection {

      [key ( true )]
   TST_Person REF family;

      [key ( true )]
   TST_Person REF member;

};
"""


OK = True     # mark tests OK when they execute correctly
RUN = True    # Mark OK = False and current test case being created RUN
FAIL = False  # Any test currently FAILING or not tested yet

# pylint: enable=line-too-long
TEST_CASES = [
    # desc - Description of test
    # inputs - String, or list of args or dict of 'env', 'args', 'general',
    #          and 'stdin'. See See CLITestsBase.command_test()  for
    #          detailed documentation
    # exp_response - Dictionary of expected responses (stdout, stderr, rc) and
    #                test definition (test: <testname>).
    #                See CLITestsBase.command_test() for detailed documentation.
    # mock - None or name of files (mof or .py),
    # condition - If True, the test is executed,  Otherwise it is skipped.

    ['Verify class command --help response',
     ['--help'],
     {'stdout': CLASS_HELP_LINES,
      'test': 'innows'},
     None, OK],

    ['Verify class command -h response',
     ['-h'],
     {'stdout': CLASS_HELP_LINES,
      'test': 'innows'},
     None, OK],

    #
    # Enumerate command and its options
    #
    ['Verify class command enumerate --help response',
     ['enumerate', '--help'],
     {'stdout': CLASS_ENUMERATE_HELP_LINES,
      'test': 'innows'},
     None, OK],

    ['Verify class command enumerate -h response',
     ['enumerate', '-h'],
     {'stdout': CLASS_ENUMERATE_HELP_LINES,
      'test': 'innows'},
     None, OK],

    ['Verify class command enumerate CIM_Foo',
     ['enumerate', 'CIM_Foo'],
     {'stdout': ['[Description ( "Subclass of CIM_Foo" )]', ],
      'test': 'regex'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command enumerate CIM_Foo --lo',
     ['enumerate', 'CIM_Foo', '--lo'],
     {'stdout':
      '   [Description ( "Subclass of CIM_Foo" )]',
      'test': 'startswith'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command enumerate CIM_Foo --lo',
     ['enumerate', 'CIM_Foo', '--local-only'],
     {'stdout':
      '   [Description ( "Subclass of CIM_Foo" )]',
      'test': 'startswith'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command enumerate CIM_Foo_sub',
     ['enumerate', 'CIM_Foo_sub'],
     {'stdout': CIMFOO_SUB_SUB,
      'test': 'linesnows'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command enumerate CIM_Foo local-only',
     ['enumerate', 'CIM_Foo', '--local-only'],
     {'stdout':
      '   [Description ( "Subclass of CIM_Foo" )]',
      'test': 'startswith'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command enumerate CIM_Foo -no-qualifiers',
     ['enumerate', 'CIM_Foo_sub', '--no-qualifiers'],
     {'stdout': CIMFOO_SUB_SUB_NO_QUALS,
      'test': 'linesnows'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command enumerate CIM_Foo --di',
     ['enumerate', 'CIM_Foo', '--di'],
     {'stdout':
      '   [Description ( "Subclass of CIM_Foo" )]',
      'test': 'startswith'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command enumerate CIM_Foo --deep-inheritance',
     ['enumerate', 'CIM_Foo', '--deep-inheritance'],
     {'stdout':
      '   [Description ( "Subclass of CIM_Foo" )]',
      'test': 'startswith'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command enumerate CIM_Foo --ico',
     ['enumerate', 'CIM_Foo', '--ico'],
     {'stdout':
      '   [Description ( "Subclass of CIM_Foo" )]',
      'test': 'startswith'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command enumerate CIM_Foo --include-classorigin',
     ['enumerate', 'CIM_Foo', '--include-classorigin'],
     {'stdout':
      '   [Description ( "Subclass of CIM_Foo" )]',
      'test': 'startswith'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command enumerate CIM_Foo --no names only',
     ['enumerate', 'CIM_Foo', '--no'],
     {'stdout': ['CIM_Foo', 'CIM_Foo_sub', 'CIM_Foo_sub2'],
      'test': 'in'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command enumerate CIM_Foo --names only',
     ['enumerate', 'CIM_Foo', '--names-only'],
     {'stdout': ['CIM_Foo', 'CIM_Foo_sub', 'CIM_Foo_sub2'],
      'test': 'in'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command enumerate --no names only - table',
     {'args': ['enumerate', '--no'],
      'general': ['--output-format', 'table']},
     {'stdout': """Classnames:
+--------------+
| Class Name   |
|--------------|
| CIM_Foo      |
+--------------+
""",
      'test': 'in'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command enumerate CIM_Foo --no names only - table',
     {'args': ['enumerate', 'CIM_Foo', '--no'],
      'general': ['--output-format', 'table']},
     {'stdout': """Classnames:
+--------------+
| Class Name   |
|--------------|
| CIM_Foo_sub  |
| CIM_Foo_sub2 |
+--------------+
""",
      'test': 'in'},
     SIMPLE_MOCK_FILE, OK],


    ['Verify class command enumerate CIM_Foo --names-only',
     ['enumerate', 'CIM_Foo', '--names-only'],
     {'stdout': ['CIM_Foo', 'CIM_Foo_sub', 'CIM_Foo_sub2'],
      'test': 'in'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command enumerate CIM_Foo summary',
     ['enumerate', 'CIM_Foo', '-s'],
     {'stdout': ['2 CIMClass(s) returned'],
      'test': 'in'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command enumerate CIM_Foo summary',
     ['enumerate', 'CIM_Foo', '--summary'],
     {'stdout': ['2 CIMClass(s) returned'],
      'test': 'linesnows'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command enumerate CIM_Foo summary table output',
     {'args': ['enumerate', 'CIM_Foo', '--summary'],
      'general': ['--output-format', 'table']},
     {'stdout': ["""Summary of CIMClass returned
+---------+------------+
|   Count | CIM Type   |
|---------+------------|
|       2 | CIMClass   |
+---------+------------+
"""],
      'test': 'linesnows'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command enumerate CIM_Foo names and --di --no',
     ['enumerate', 'CIM_Foo', '--di', '--no'],
     {'stdout': ['CIM_Foo_sub', 'CIM_Foo_sub2', 'CIM_Foo_sub_sub'],
      'test': 'in'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command enumerate CIM_Foo names and --deep-inheritance '
      '--names-only',
     ['enumerate', 'CIM_Foo', '--names-only', '--deep-inheritance'],
     {'stdout': ['CIM_Foo_sub', 'CIM_Foo_sub2', 'CIM_Foo_sub_sub'],
      'test': 'in'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command enumerate CIM_Foo include qualifiers',
     ['enumerate', 'CIM_Foo'],
     {'stdout': ['Key ( true )', '[Description (', 'class CIM_Foo'],
      'test': 'in'},
     SIMPLE_MOCK_FILE, OK],


    ['Verify class command get with xml output format).',
     {'args': ['enumerate'],
      'general': ['--output-format', 'repr']},
     {'stdout': [r"CIMClass\(classname='CIM_Foo', superclass=None,",
                 r"'InstanceID': CIMProperty\(name='InstanceID', value=None,"],
      'test': 'regex'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command get with repr output format).',
     {'args': ['enumerate'],
      'general': ['--output-format', 'txt']},
     {'stdout': ["CIMClass(classname='CIM_Foo', ...)"],
      'test': 'lines'},
     SIMPLE_MOCK_FILE, OK],


    ['Verify class command get with repr output format).',
     {'args': ['enumerate'],
      'general': ['--output-format', 'xml']},
     {'stdout': ['<CLASS NAME="CIM_Foo">',
                 '<PROPERTY NAME="InstanceID"',
                 '<PROPERTY NAME="IntegerProp" PROPAGATED="false"',
                 '<METHOD NAME="DeleteNothing" PROPAGATED="false"'],
      'test': 'regex'},
     SIMPLE_MOCK_FILE, OK],
    #
    # Enumerate errors
    #

    ['Verify class command enumerate nonexistent class name',
     ['enumerate', 'CIM_FClassDoesNotExist'],
     {'stderr': ['CIMError', 'CIM_ERR_INVALID_CLASS'],
      'rc': 1,
      'test': 'regex'},
     SIMPLE_MOCK_FILE, OK],

    #
    # Test class get
    #
    ['Verify class command get --help response',
     ['get', '--help'],
     {'stdout': CLASS_GET_HELP_LINES,
      'test': 'innows'},
     None, OK],

    ['Verify class command get -h response',
     ['get', '-h'],
     {'stdout': CLASS_GET_HELP_LINES,
      'test': 'innows'},
     None, OK],

    # command get local-only option
    ['Verify class command get not local-only. Tests for property names',
     ['get', 'CIM_Foo_sub2'],
     {'stdout': ['string cimfoo_sub2;', 'InstanceID', 'IntegerProp', 'Fuzzy',
                 'Key ( true )', 'IN ( false )'],
      'test': 'in'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command get local-only(--lo)).',
     ['get', 'CIM_Foo_sub2', '--lo'],
     {'stdout': ['class CIM_Foo_sub2 : CIM_Foo {',
                 '',
                 '   string cimfoo_sub2;',
                 '',
                 '};', '', ],
      'test': 'patterns'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command get local-only. Tests whole response',
     ['get', 'CIM_Foo_sub2', '--local-only'],
     {'stdout': ['class CIM_Foo_sub2 : CIM_Foo {',
                 '',
                 '   string cimfoo_sub2;',
                 '',
                 '};', '', ],
      'test': 'patterns'},
     SIMPLE_MOCK_FILE, OK],

    # includequalifiers. Test the flag that excludes qualifiers
    ['Verify class command get without qualifiers. Tests whole response',
     ['get', 'CIM_Foo_sub2', '--nq'],
     {'stdout': ['class CIM_Foo_sub2 : CIM_Foo {',
                 '',
                 '   string cimfoo_sub2;',
                 '',
                 '   string InstanceID;',
                 '',
                 '   uint32 IntegerProp;',
                 '',
                 '   uint32 Fuzzy(',
                 '      string TestInOutParameter,',
                 '      CIM_Foo REF TestRef,',
                 '      string OutputParam,',
                 '      uint32 OutputRtnValue);',
                 '',
                 '   uint32 DeleteNothing();',
                 '',
                 '};',
                 '', ],
      'test': 'lines'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command get without qualifiers. Tests whole response',
     ['get', 'CIM_Foo_sub2', '--no-qualifiers'],
     {'stdout': ['class CIM_Foo_sub2 : CIM_Foo {',
                 '',
                 '   string cimfoo_sub2;',
                 '',
                 '   string InstanceID;',
                 '',
                 '   uint32 IntegerProp;',
                 '',
                 '   uint32 Fuzzy(',
                 '      string TestInOutParameter,',
                 '      CIM_Foo REF TestRef,',
                 '      string OutputParam,',
                 '      uint32 OutputRtnValue);',
                 '',
                 '   uint32 DeleteNothing();',
                 '',
                 '};',
                 '', ],
      'test': 'lines'},
     SIMPLE_MOCK_FILE, OK],

    # pylint: disable=line-too-long
    ['Verify class command get with propertylist. Tests whole response',
     ['get', 'CIM_Foo_sub2', '--pl', 'InstanceID'],
     {'stdout': ['class CIM_Foo_sub2 : CIM_Foo {', '',
                 '      [Key ( true ),',
                 '       Description ( "This is key property." )]', ''
                 '   string InstanceID;', '',
                 '      [Description ( "Method with in and out parameters" )'
                 ']',
                 '   uint32 Fuzzy(',
                 '         [IN ( true ),',
                 '          OUT ( true ),',
                 '          Description ( "Define data to be returned in output parameter" )]',  # noqa: E501
                 '      string TestInOutParameter,',
                 '         [IN ( true ),',
                 '          OUT ( true ),',
                 '          Description ( "Test of ref in/out parameter" )]',
                 '      CIM_Foo REF TestRef,',
                 '         [IN ( false ),',
                 '          OUT ( true ),',
                 '          Description ( "Rtns method name if exists on input" )]',  # noqa: E501
                 '      string OutputParam,',
                 '         [IN ( true ),',
                 '          Description ( "Defines return value if provided." )]',  # noqa: E501
                 '      uint32 OutputRtnValue);',
                 '',
                 '      [Description ( "Method with no Parameters" )]',
                 '   uint32 DeleteNothing();',
                 '',
                 '};',
                 '', ],
      'test': 'lines'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command get with empty propertylist. Tests whole '
     'response',
     ['get', 'CIM_Foo_sub2', '--pl', '""'],
     {'stdout': ['class CIM_Foo_sub2 : CIM_Foo {', '',
                 '      [Description ( "Method with in and out parameters" )'
                 ']',
                 '   uint32 Fuzzy(',
                 '         [IN ( true ),',
                 '          OUT ( true ),',
                 '          Description ( "Define data to be returned in output parameter" )]',  # noqa: E501
                 '      string TestInOutParameter,',
                 '         [IN ( true ),',
                 '          OUT ( true ),',
                 '          Description ( "Test of ref in/out parameter" )]',
                 '      CIM_Foo REF TestRef,',
                 '         [IN ( false ),',
                 '          OUT ( true ),',
                 '          Description ( "Rtns method name if exists on input" )]',  # noqa: E501
                 '      string OutputParam,',
                 '         [IN ( true ),',
                 '          Description ( "Defines return value if provided." )]',  # noqa: E501
                 '      uint32 OutputRtnValue);',
                 '',
                 '      [Description ( "Method with no Parameters" )]',
                 '   uint32 DeleteNothing();',
                 '',
                 '};',
                 '', ],
      'test': 'lines'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command get with xml output format).',
     {'args': ['get', 'CIM_Foo'],
      'general': ['--output-format', 'repr']},
     {'stdout': [r"CIMClass\(classname='CIM_Foo', superclass=None,",
                 r"'InstanceID': CIMProperty\(name='InstanceID', value=None,"],
      'test': 'regex'},
     SIMPLE_MOCK_FILE, OK],


    ['Verify class command get with repr output format).',
     {'args': ['get', 'CIM_Foo'],
      'general': ['--output-format', 'txt']},
     {'stdout': ["CIMClass(classname='CIM_Foo', ...)"],
      'test': 'lines'},
     SIMPLE_MOCK_FILE, OK],


    ['Verify class command get with repr output format).',
     {'args': ['get', 'CIM_Foo'],
      'general': ['--output-format', 'xml']},
     {'stdout': ['<CLASS NAME="CIM_Foo">',
                 '<PROPERTY NAME="InstanceID"',
                 '<PROPERTY NAME="IntegerProp" PROPAGATED="false"',
                 '<METHOD NAME="DeleteNothing" PROPAGATED="false"'],
      'test': 'regex'},
     SIMPLE_MOCK_FILE, OK],

    # pylint: enable=line-too-long
    # TODO include class origin. TODO not returning class origin correctly.
    ['Verify class command get with propertylist and classorigin,',
     ['get', 'CIM_Foo_sub2', '--pl', 'InstanceID', '-c'],
     {'stdout': ['class CIM_Foo_sub2 : CIM_Foo {', '',
                 '      [Key ( true ),',
                 '       Description ( "This is key property." )]', ''
                 '   string InstanceID;', '',
                 '      [Description ( "Method with in and out parameters" )'
                 ']',
                 '   uint32 Fuzzy(',
                 '         [IN ( true ),',
                 '          Description ( "FuzzyMethod Param" )]',
                 '      string FuzzyParameter,',
                 '         [IN ( true ),',
                 '          OUT ( true ),',
                 '          Description ( "Test of ref in/out parameter" )]',
                 '      CIM_Foo REF Foo,',
                 '         [IN ( false ),',
                 '          OUT ( true ),',
                 '          Description ( "TestMethod Param" )]',
                 '      string OutputParam,',
                 '',
                 '      [Description ( "Method with no Parameters" )]',
                 '',
                 '   uint32 DeleteNothing();', '',
                 '};', '', ],
      'test': 'lines'},
     SIMPLE_MOCK_FILE, FAIL],

    # get command errors

    ['Verify class command get invalid classname',
     ['get', 'CIM_Argh'],
     {'stderr': ['CIMError', 'CIM_ERR_NOT_FOUND', '6'],
      'rc': 1,
      'test': 'regex'},

     SIMPLE_MOCK_FILE, OK],
    ['Verify class command get invalid namespace',
     ['get', 'CIM_Foo', '--namespace', 'Argh'],
     {'stderr': ['CIMError', 'CIM_ERR_INVALID_NAMESPACE', '3'],
      'rc': 1,
      'test': 'regex'},
     SIMPLE_MOCK_FILE, OK],

    #
    # find command
    #
    ['Verify class command find --help response',
     ['find', '--help'],
     {'stdout': CLASS_FIND_HELP_LINES,
      'test': 'innows'},
     None, OK],

    ['Verify class command find -h response',
     ['find', '-h'],
     {'stdout': CLASS_FIND_HELP_LINES,
      'test': 'innows'},
     None, OK],

    ['Verify class command find simple name in all namespaces',
     ['find', 'CIM_*'],
     {'stdout': ["  root/cimv2:CIM_Foo",
                 "  root/cimv2:CIM_Foo_sub",
                 "  root/cimv2:CIM_Foo_sub2",
                 "  root/cimv2:CIM_Foo_sub_sub"],
      'test': 'in'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command find simple name in all namespaces wo case',
     ['find', 'cim_*'],
     {'stdout': ["  root/cimv2:CIM_Foo",
                 "  root/cimv2:CIM_Foo_sub",
                 "  root/cimv2:CIM_Foo_sub2",
                 "  root/cimv2:CIM_Foo_sub_sub"],
      'test': 'in'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command find simple name in all namespaces lead wc',
     ['find', '*sub_sub*'],
     {'stdout': ["  root/cimv2:CIM_Foo_sub_sub"],
      'test': 'in'},
     SIMPLE_MOCK_FILE, OK],


    ['Verify class command find simple name in all namespaces wo case',
     ['find', '*sub_su?*'],
     {'stdout': ["  root/cimv2:CIM_Foo_sub_sub"],
      'test': 'in'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command find simple name in known namespace',
     ['find', 'CIM_*', '-n', 'root/cimv2'],
     {'stdout': ["  root/cimv2:CIM_Foo",
                 "  root/cimv2:CIM_Foo_sub",
                 "  root/cimv2:CIM_Foo_sub2",
                 "  root/cimv2:CIM_Foo_sub_sub"],
      'test': 'lines'},
     SIMPLE_MOCK_FILE, OK],


    ['Verify class command find name in known namespace -o grid',
     {'general': ['-o', 'grid'],
      'args': ['find', 'CIM_*', '-n', 'root/cimv2']},
     {'stdout': ['Find class CIM_*',
                 '+-------------+-----------------+',
                 '| Namespace   | Classname       |',
                 '+=============+=================+',
                 '| root/cimv2  | CIM_Foo         |',
                 '+-------------+-----------------+',
                 '| root/cimv2  | CIM_Foo_sub     |',
                 '+-------------+-----------------+',
                 '| root/cimv2  | CIM_Foo_sub2    |',
                 '+-------------+-----------------+',
                 '| root/cimv2  | CIM_Foo_sub_sub |',
                 '+-------------+-----------------+', ],
      'test': 'lines'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command verify nothing found for BLAH_ regex',
     ['find', 'BLAH_*', '-n', 'root/cimv2'],
     {'stdout': "",
      'test': 'lines'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command find simple name in known namespace with wildcard',
     ['find', '*sub2', '-n', 'root/cimv2'],
     {'stdout': "  root/cimv2:CIM_Foo_sub2",
      'test': 'lines'},
     SIMPLE_MOCK_FILE, OK],

    #
    # command "class delete"
    #
    ['Verify class command delete --help response',
     ['delete', '--help'],
     {'stdout': CLASS_DELETE_HELP_LINES,
      'test': 'innows'},
     None, OK],

    ['Verify class command delete -h response',
     ['delete', '-h'],
     {'stdout': CLASS_DELETE_HELP_LINES,
      'test': 'innows'},
     None, OK],

    # Class delete successful
    ['Verify class command delete successful with no subclasses, --force',
     ['delete', 'CIM_Foo_sub_sub', '--force'],
     {'stdout': 'CIM_Foo_sub_sub delete successful',
      'test': 'in'},
     [SIMPLE_MOCK_FILE_EXT], OK],

    ['Verify class command delete fail instances exist',
     ['delete', 'CIM_Foo_sub_sub'],
     {'stdout': 'CIM_Foo_sub_sub delete successful',
      'rc': 0,
      'test': 'in'},
     SIMPLE_MOCK_FILE, OK],

    # Class delete errors
    ['Verify class command delete no classname',
     ['delete'],
     {'stderr': 'Error: Missing argument "CLASSNAME"',
      'rc': 2,
      'test': 'in'},
     SIMPLE_MOCK_FILE, OK],

    # class delete error tests

    ['Verify class command delete fail subclasses exist',
     ['delete', 'CIM_Foo', '--force'],
     {'stderr': 'Error: Delete rejected; subclasses exist',
      'rc': 1,
      'test': 'in'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command delete no classname fails',
     ['delete', ],
     {'stderr': 'Missing argument "CLASSNAME"',
      'rc': 2,
      'test': 'regex'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command delete nonexistent classname fails',
     ['delete', 'Argh'],
     {'stderr': ['CIMError', 'CIM_ERR_INVALID_CLASS', '5'],
      'rc': 1,
      'test': 'regex'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command delete fails, instances exist',
     ['delete', 'CIM_Foo_sub_sub'],
     {'stderr': 'Error: Delete rejected; instances exist',
      'rc': 1,
      'test': 'in'},
     [SIMPLE_MOCK_FILE_EXT], OK],

    #
    # command "class tree"
    #
    ['Verify class command tree --help response',
     ['tree', '--help'],
     {'stdout': CLASS_TREE_HELP_LINES,
      'test': 'innows'},
     None, OK],

    ['Verify class command tree -h response',
     ['tree', '-h'],
     {'stdout': CLASS_TREE_HELP_LINES,
      'test': 'innows'},
     None, OK],

    # Order inconsistent on output. so we just test that some of the lines are

    # in the output
    # root
    #  +-- CIM_Foo
    #      +-- CIM_Foo_sub2
    #      +-- CIM_Foo_sub
    #          +-- CIM_Foo_sub_sub
    # or
    # root
    #  +-- CIM_Foo
    #      +-- CIM_Foo_sub
    #      |   +-- CIM_Foo_sub_sub
    #      +-- CIM_Foo_sub2
    ['Verify class command tree top down. Order uncertain.',
     ['tree'],
     {'stdout': [r'root',
                 r' \+-- CIM_Foo',
                 r'     \+-- CIM_Foo_sub2',
                 r'     \+-- CIM_Foo_sub',
                 r'[| ]*\+-- CIM_Foo_sub_sub', ],  # account for ordering issue
      'test': 'regex'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command tree top down starting at defined class ',
     ['tree', 'CIM_Foo_sub'],
     {'stdout': ['CIM_Foo_sub',
                 ' +-- CIM_Foo_sub_sub', ],
      'test': 'lines'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command tree bottom up. Ordering guaranteed here',
     ['tree', '-s', 'CIM_Foo_sub_sub'],
     {'stdout': ['root',
                 ' +-- CIM_Foo',
                 '     +-- CIM_Foo_sub',
                 '         +-- CIM_Foo_sub_sub', ],
      'test': 'lines'},
     SIMPLE_MOCK_FILE, OK],

    # class tree' error tests
    ['Verify class command tree with invalid class',
     ['tree', '-s', 'CIM_Foo_subx'],
     {'stderr': ['CIMError:'],
      'rc': 1,
      'test': 'regex'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify class command tree with superclass option, no class',
     ['tree', '-s'],
     {'stderr': ['Error: CLASSNAME argument required for --superclasses '
                 'option'],
      'rc': 1,
      'test': 'regex'},
     SIMPLE_MOCK_FILE, OK],

    #
    # associators command tests
    #
    #
    ['Verify class command associators --help response',
     ['associators', '--help'],
     {'stdout': CLASS_ASSOCIATORS_HELP_LINES,
      'test': 'innows'},
     None, OK],

    ['Verify class command associators -h response',
     ['associators', '-h'],
     {'stdout': CLASS_ASSOCIATORS_HELP_LINES,
      'test': 'innows'},
     None, OK],

    ['Verify class command associators simple request,',
     ['associators', 'TST_Person', ],
     {'stdout': ['//FakedUrl/root/cimv2:TST_Person',
                 'class TST_Person {',
                 '',
                 '      [Key ( true ),',
                 '       Description ( "This is key prop" )]',
                 '   string name;',
                 '',
                 '   string extraProperty = "defaultvalue";',
                 '',
                 '};',
                 ''],
      'test': 'lines'},
     SIMPLE_ASSOC_MOCK_FILE, OK],


    ['Verify class command associators simple request names only,',
     ['associators', 'TST_Person', '--names-only'],
     {'stdout': ['//FakedUrl/root/cimv2:TST_Person'],
      'test': 'lines'},
     SIMPLE_ASSOC_MOCK_FILE, OK],

    ['Verify class command associators simple request, one parameter',
     ['associators', 'TST_Person', '--ac', 'TST_MemberOfFamilyCollection'],
     {'stdout': ['//FakedUrl/root/cimv2:TST_Person',
                 'class TST_Person {',
                 '',
                 '      [Key ( true ),',
                 '       Description ( "This is key prop" )]',
                 '   string name;',
                 '',
                 '   string extraProperty = "defaultvalue";',
                 '',
                 '};',
                 ''],
      'test': 'lines'},
     SIMPLE_ASSOC_MOCK_FILE, OK],

    ['Verify class command associators request, all filters long',
     ['associators', 'TST_Person',
      '--assoc-class', 'TST_MemberOfFamilyCollection',
      '--role', 'member',
      '--result-role', 'family',
      '--result-class', 'TST_Person'],
     {'stdout': ['//FakedUrl/root/cimv2:TST_Person',
                 'class TST_Person {',
                 '',
                 '      [Key ( true ),',
                 '       Description ( "This is key prop" )]',
                 '   string name;',
                 '',
                 '   string extraProperty = "defaultvalue";',
                 '',
                 '};',
                 ''],
      'test': 'lines'},
     SIMPLE_ASSOC_MOCK_FILE, OK],

    ['Verify class command associators request, all filters short',
     ['associators', 'TST_Person',
      '--ac', 'TST_MemberOfFamilyCollection',
      '-r', 'member',
      '--rr', 'family',
      '--rc', 'TST_Person'],
     {'stdout': ['//FakedUrl/root/cimv2:TST_Person',
                 'class TST_Person {',
                 '',
                 '      [Key ( true ),',
                 '       Description ( "This is key prop" )]',
                 '   string name;',
                 '',
                 '   string extraProperty = "defaultvalue";',
                 '',
                 '};',
                 ''],
      'test': 'lines'},
     SIMPLE_ASSOC_MOCK_FILE, OK],

    ['Verify class command associators request, all filters short,  -a '
     'does not pass test',
     ['associators', 'TST_Person',
      '--ac', 'TST_MemberOfFamilyCollectionx',
      '-r', 'member',
      '--rr', 'family',
      '--rc', 'TST_Person'],
     {'stdout': [],
      'test': 'lines'},
     SIMPLE_ASSOC_MOCK_FILE, OK],

    ['Verify class command associators request, all filters short,  -r '
     'does not pass test',
     ['associators', 'TST_Person',
      '--ac', 'TST_MemberOfFamilyCollection',
      '-r', 'memberx',
      '--rr', 'family',
      '--rc', 'TST_Person'],
     {'stdout': [],
      'test': 'lines'},
     SIMPLE_ASSOC_MOCK_FILE, OK],

    ['Verify class command associators request, all filters short,  -rr '
     'does not pass test',
     ['associators', 'TST_Person',
      '--ac', 'TST_MemberOfFamilyCollection',
      '-r', 'member',
      '--rr', 'familyx',
      '--rc', 'TST_Person'],
     {'stdout': [],
      'test': 'lines'},
     SIMPLE_ASSOC_MOCK_FILE, OK],

    ['Verify class command associators request, all filters short,  --rc '
     'does not pass test',
     ['associators', 'TST_Person',
      '--ac', 'TST_MemberOfFamilyCollection',
      '-r', 'member',
      '--rr', 'family',
      '--rc', 'TST_Personx'],
     {'stdout': [],
      'test': 'lines'},
     SIMPLE_ASSOC_MOCK_FILE, OK],

    ['Verify class command associators request, all filters long '
     'does not pass test',
     ['associators', 'TST_Person',
      '--assoc-class', 'TST_MemberOfFamilyCollection',
      '--role', 'member',
      '--result-role', 'family',
      '--result-class', 'TST_Personx'],
     {'stdout': [],
      'test': 'lines'},
     SIMPLE_ASSOC_MOCK_FILE, OK],

    # Associator errors

    ['Verify class command associators no CLASSNAME',
     ['associators'],
     {'stderr': ['Error: Missing argument "CLASSNAME".', ],
      'rc': 2,
      'test': 'in'},
     None, OK],

    ['Verify class command associators non-existent CLASSNAME rtns empty',
     ['associators', 'CIM_Nonexistentclass'],
     {'stdout': "",
      'rc': 0,
      'test': 'regex'},
     SIMPLE_ASSOC_MOCK_FILE, OK],

    ['Verify class command associators non-existent namespace fails',
     ['associators', 'TST_Person', '--namespace', 'blah'],
     {'stderr': ['CIMError', 'CIM_ERR_INVALID_NAMESPACE'],
      'rc': 1,
      'test': 'regex'},
     SIMPLE_ASSOC_MOCK_FILE, OK],

    #
    # references command tests
    #
    ['Verify class command references --help response',
     ['references', '--help'],
     {'stdout': CLASS_REFERENCES_HELP_LINES,
      'test': 'innows'},
     None, OK],

    ['Verify class command references -h response',
     ['references', '-h'],
     {'stdout': CLASS_REFERENCES_HELP_LINES,
      'test': 'innows'},
     None, OK],

    ['Verify class command references simple request',
     ['references', 'TST_Person'],
     {'stdout': REFERENCES_CLASS_RTN_QUALS1,
      'test': 'linesnows'},
     SIMPLE_ASSOC_MOCK_FILE, OK],

    ['Verify class command references simple request -o',
     ['references', 'TST_Person', '--no'],
     {'stdout': ['//FakedUrl/root/cimv2:TST_Lineage',
                 '//FakedUrl/root/cimv2:TST_MemberOfFamilyCollection'],
      'test': 'linesnows'},
     SIMPLE_ASSOC_MOCK_FILE, OK],

    ['Verify class command references request, all filters long',
     ['references', 'TST_Person',
      '--role', 'member',
      '--result-class', 'TST_MemberOfFamilyCollection'],
     {'stdout': REFERENCES_CLASS_RTN_QUALS2,
      'test': 'linesnows'},
     SIMPLE_ASSOC_MOCK_FILE, OK],


    ['Verify class command references request, filters short',
     ['references', 'TST_Person',
      '-r', 'member',
      '--rc', 'TST_MemberOfFamilyCollection'],
     {'stdout': REFERENCES_CLASS_RTN_QUALS2,
      'test': 'linesnows'},
     SIMPLE_ASSOC_MOCK_FILE, OK],

    # Reference errors

    ['Verify class command references no CLASSNAME',
     ['references'],
     {'stderr': ['Error: Missing argument "CLASSNAME".', ],
      'rc': 2,
      'test': 'in'},
     None, OK],

    ['Verify class command references non-existent CLASSNAME rtns empty',
     ['references', 'CIM_Nonexistentclass'],
     {'stdout': "",
      'rc': 0,
      'test': 'regex'},
     SIMPLE_ASSOC_MOCK_FILE, OK],

    ['Verify class command references non-existent namespace fails',
     ['references', 'TST_Person', '--namespace', 'blah'],
     {'stderr': ['CIMError', 'CIM_ERR_INVALID_NAMESPACE'],
      'rc': 1,
      'test': 'regex'},
     SIMPLE_ASSOC_MOCK_FILE, OK],

    #
    # invokemethod command tests
    #
    ['Verify class command invokemethod --help response',
     ['invokemethod', '--help'],
     {'stdout': CLASS_INVOKEMETHOD_HELP_LINES,
      'test': 'innows'},
     None, OK],

    ['Verify class command invokemethod -h response',
     ['invokemethod', '-h'],
     {'stdout': CLASS_INVOKEMETHOD_HELP_LINES,
      'test': 'innows'},
     None, OK],

    #
    #  class invokemethod command without parameters
    #
    ['Verify class command invokemethod',
     ['invokemethod', 'CIM_Foo', 'Fuzzy'],
     {'stdout': ["ReturnValue=0"],
      'rc': 0,
      'test': 'lines'},
     [SIMPLE_MOCK_FILE, INVOKE_METHOD_MOCK_FILE], OK],

    ['Verify class command invokemethod',
     ['invokemethod', 'CIM_Foo', 'Fuzzy', '-p', 'TestInOutParameter="blah"'],
     {'stdout': ['ReturnValue=0',
                 'TestInOutParameter=', 'blah'],
      'rc': 0,
      'test': 'innows'},
     [SIMPLE_MOCK_FILE, INVOKE_METHOD_MOCK_FILE], OK],

    ['Verify class command invokemethod fails Invalid Class',
     ['invokemethod', 'CIM_Foox', 'Fuzzy', '-p', 'TestInOutParameter="blah"'],
     {'stderr': ["Error: CIMError: 6"],
      'rc': 1,
      'test': 'in'},
     [SIMPLE_MOCK_FILE, INVOKE_METHOD_MOCK_FILE], OK],

    ['Verify class command invokemethod fails Invalid Method',
     ['invokemethod', 'CIM_Foo', 'Fuzzyx', '-p', 'TestInOutParameter=blah'],
     {'stderr': ["Error: ClickException: .* CIM_Foo .* Fuzzyx"],
      'rc': 1,
      'test': 'regex'},
     [SIMPLE_MOCK_FILE, INVOKE_METHOD_MOCK_FILE], OK],


    ['Verify class command invokemethod fails Method not registered',
     ['invokemethod', 'CIM_Foo', 'Fuzzy'],
     {'stderr': ["CIMError: 17"],
      'rc': 1,
      'test': 'innows'},
     [SIMPLE_MOCK_FILE], OK],

    ['Verify  --timestats gets stats output. Cannot test '
     'with lines because execution time is variable.',
     {'args': ['get', 'CIM_Foo'],
      'general': ['--timestats']},
     {'stdout': ['class CIM_Foo {',
                 'string InstanceID;',
                 'Count    Exc    Time    ReqLen    ReplyLen  Operation',
                 '      1      0',
                 '0           0  GetClass'],
      'rc': 0,
      'test': 'regex'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify single command with stdin works',
     {'stdin': ['class get -h']},
     {'stdout': ['Usage: pywbemcli  class get '],
      'rc': 0,
      'test': 'regex'},
     None, OK],

    ['Verify multiple commands with stdin work',
     {'stdin': ['class get -h', 'class enumerate -h']},
     {'stdout': ['Usage: pywbemcli  class enumerate ',
                 'Usage: pywbemcli  class get '],
      'rc': 0,
      'test': 'regex'},
     None, OK],
]

# TODO command class delete. Extend this test to use stdin (delete, test)
# namespace
# TODO: add test for  errors: class invalid, namespace invalid
# other tests.  Test local-only on top level


# TODO the following two test classes should be removed as I believe they
# are redundant.


class TestSubcmdClass(CLITestsBase):
    """
    Test all of the class command variations.
    """
    command_group = 'class'

    @pytest.mark.parametrize(
        "desc, inputs, exp_response, mock, condition",
        TEST_CASES)
    def test_class(self, desc, inputs, exp_response, mock, condition):
        """
        Common test method for those commands and options in the
        class command that can be tested.  This includes:

          * Subcommands like help that do not require access to a server

          * Subcommands that can be tested with a single execution of a
            pywbemcli command.
        """
        self.command_test(desc, self.command_group, inputs, exp_response,
                          mock, condition, verbose=False)


class TestClassGeneral(object):
    # pylint: disable=too-few-public-methods, useless-object-inheritance
    """
    Test class using pytest for the commands of the class command
    """
    # @pytest.mark.skip(reason="Unfinished test")
    def test_class_error_no_server(self):  # pylint: disable=no-self-use
        """Test 'pywbemcli ... class getclass' when no host is provided

        This test runs against a real url so we set timeout to the mininum
        to minimize test time since the expected result is a timeout exception.
        """

        # Invoke the command to be tested
        rc, stdout, stderr = execute_pywbemcli(['-s', 'http://fred', '-t', '1',
                                                'class', 'get', 'CIM_blah'])

        assert_rc(1, rc, stdout, stderr)

        assert stdout == ""
        assert stderr.startswith(
            "Error: ConnectionError"), \
            "stderr={!r}".format(stderr)


class TestClassEnumerate(object):
    # pylint: disable=too-few-public-methods, useless-object-inheritance
    """
    Test the options of the pywbemcli class enumerate' command
    """
    @pytest.mark.parametrize(
        "desc, tst_args, exp_result_start, exp_result_end",
        [
            [
                ["No extra parameters"],
                [],
                '   [Description ( "Simple CIM Class" )]\n',
                None
            ],
            [
                ["Class parameter"],
                ['CIM_Foo'],
                '   [Description ( "Subclass of CIM_Foo" )]\n',
                None
            ]
        ]
    )
    # pylint: disable=unused-argument
    def test_enumerate_simple_mock(self, desc, tst_args, exp_result_start,
                                   exp_result_end):
        # pylint: disable=no-self-use
        """
        Test 'pywbemcli class enumerate command based on a simple set of
        classes defined in the file simple_mock_model.mof
        """
        mock_mof_path = os.path.join(TEST_DIR, 'simple_mock_model.mof')

        # build basic cmd line, server, mock-server def, basic enum command
        cmd_line = ['--mock-server', mock_mof_path, 'class', 'enumerate']

        if tst_args:
            cmd_line.extend(tst_args)
        rc, stdout, stderr = execute_pywbemcli(cmd_line)

        assert_rc(0, rc, stdout, stderr)
        assert stderr == ""
        assert stdout.startswith(exp_result_start)



#!/usr/bin/env python

# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for PywbemServer class methods.
"""

from __future__ import absolute_import, print_function

import unittest

from pywbemtools.pywbemcli._pywbem_server import PywbemServer

# TODO: Future - Move these tests to pytest


class PywbemServerTests(unittest.TestCase):
    """ Test the PywbemServer class """

    def test_simple(self):
        """ Create the object and test values"""

        server = 'http://localhost'
        ns = 'root/cimv2'
        user = 'Fred'
        pw = 'blah'
        svr = PywbemServer(server, ns, user=user, password=pw)

        self.assertEqual(svr.server, server)
        self.assertEqual(svr.default_namespace, ns)
        self.assertEqual(svr.user, user)
        self.assertEqual(svr.password, pw)

        print(svr)

    def test_all_parms(self):
        """Test all the PywbemServer input parameters"""
        server = 'http://localhost'
        ns = 'root/cimv2'
        user = 'Fred'
        pw = 'blah'
        timeout = 10
        verify = True
        certfile = 'mycertfile.blah'
        keyfile = 'mykeys.blah'

        svr = PywbemServer(server, ns, user=user, password=pw, timeout=timeout,
                           verify=verify, certfile=certfile,
                           keyfile=keyfile)

        self.assertEqual(svr.server, server)
        self.assertEqual(svr.default_namespace, ns)
        self.assertEqual(svr.user, user)
        self.assertEqual(svr.password, pw)
        self.assertEqual(svr.verify, verify)
        self.assertEqual(svr.certfile, certfile)
        self.assertEqual(svr.keyfile, keyfile)

    def test_all_connect(self):
        """Test all input parameters"""
        server = 'http://localhost'
        ns = 'root/cimv2'
        user = 'Fred'
        pw = 'blah'
        timeout = 10
        verify = True
        certfile = 'mycertfile.blah'
        keyfile = 'mykeys.blah'

        svr = PywbemServer(server, ns, user=user, password=pw, timeout=timeout,
                           verify=verify, certfile=certfile,
                           keyfile=keyfile)

        self.assertEqual(svr.server, server)
        self.assertEqual(svr.default_namespace, ns)
        self.assertEqual(svr.user, user)
        self.assertEqual(svr.password, pw)
        self.assertEqual(svr.verify, verify)
        self.assertEqual(svr.certfile, certfile)
        self.assertEqual(svr.keyfile, keyfile)

        # connect and test connection results
        svr.create_connection(False)
        self.assertEqual(svr.conn.url, server)
        self.assertEqual(svr.wbem_server.conn.url, svr.conn.url)

        # TODO test for errors


if __name__ == '__main__':
    unittest.main()



# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Click Command definition for the class command group which includes
commands for get, enumerate, associators, references, find, etc. of the objects
CIMClass on a WBEM server
"""
from __future__ import absolute_import

import click

from pywbem import Error, CIMClassName, CIMError, CIM_ERR_NOT_FOUND

from .pywbemcli import cli
from ._common import display_cim_objects, filter_namelist, \
    resolve_propertylist, CMD_OPTS_TXT, TABLE_FORMATS, \
    format_table, process_invokemethod
from ._common_options import add_options, propertylist_option, \
    names_only_option, include_classorigin_class_option, namespace_option,  \
    summary_option, multiple_namespaces_option

from ._displaytree import display_class_tree


#
#   Common option definitions for class group
#

# NOTE: A number of the options use double-dash as the short form.  In those
# cases, a third definition of the options without the double-dash defines
# the corresponding option name, ex. 'include_qualifiers'. It should be
# defined with underscore and not dash

no_qualifiers_class_option = [              # pylint: disable=invalid-name
    click.option('--nq', '--no-qualifiers', 'no_qualifiers', is_flag=True,
                 default=True,
                 help='Do not include qualifiers in the returned class(es). '
                      'Default: Include qualifiers.')]

deep_inheritance_class_option = [              # pylint: disable=invalid-name
    click.option('--di', '--deep-inheritance', 'deep_inheritance', is_flag=True,
                 default=False,
                 help='Include the complete subclass hierarchy of the '
                      'requested classes in the result set. '
                      'Default: Do not include subclasses.')]

local_only_class_option = [              # pylint: disable=invalid-name
    click.option('--lo', '--local-only', 'local_only', is_flag=True,
                 default=False,
                 help='Do not include superclass properties and methods in '
                      'the returned class(es). '
                      'Default: Include superclass properties and methods.')]


@cli.group('class', options_metavar=CMD_OPTS_TXT)
def class_group():
    """
    Command group for CIM classes.

    This command group defines commands to inspect classes, to invoke
    methods on classes, and to delete classes.

    Creation and modification of classes is not currently supported.

    In addition to the command-specific options shown in this help text, the
    general options (see 'pywbemcli --help') can also be specified before the
    'class' keyword.
    """
    pass  # pylint: disable=unnecessary-pass


@class_group.command('get', options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME', required=True,)
@add_options(local_only_class_option)
@add_options(no_qualifiers_class_option)
@add_options(include_classorigin_class_option)
@add_options(propertylist_option)
@add_options(namespace_option)
@click.pass_obj
def class_get(context, classname, **options):
    """
    Get a class.

    Get a CIM class (CLASSNAME argument) in a CIM namespace (--namespace
    option). If no namespace was specified, the default namespace of the
    connection is used.

    The --local-only, --include-classorigin, --no-qualifiers, and
    --propertylist options determine which parts are included in each retrieved
    class.

    In the output, the class will be formatted as defined by the
    --output-format general option. Table formats are replaced with MOF
    format.

    Example:

      pywbemcli -n myconn class get CIM_Foo -n interop
    """
    context.execute_cmd(lambda: cmd_class_get(context, classname, options))


@class_group.command('delete', options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME', required=True,)
@click.option('-f', '--force', is_flag=True, default=False,
              help='Delete any instances of the class as well. '
                   'Some servers may still reject the class deletion. '
                   'Default: Reject command if the class has any instances.')
@add_options(namespace_option)
@click.pass_obj
def class_delete(context, classname, **options):
    """
    Delete a class.

    Delete a CIM class (CLASSNAME argument) in a CIM namespace (--namespace
    option). If no namespace was specified, the default namespace of the
    connection is used.

    If the class has subclasses, the command is rejected.

    If the class has instances, the command is rejected, unless the --force
    option was specified, in which case the instances are also deleted.

    WARNING: Deleting classes can cause damage to the server: It can impact
    instance providers and other components in the server. Use this
    command with caution.

    Many WBEM servers may not allow this operation or may severely limit the
    conditions under which a class can be deleted from the server.

    Example:

      pywbemcli -n myconn class delete CIM_Foo -n interop
    """
    context.execute_cmd(lambda: cmd_class_delete(context, classname, options))


@class_group.command('invokemethod', options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME', required=True,)
@click.argument('methodname', type=str, metavar='METHODNAME', required=True)
@click.option('-p', '--parameter', type=str, metavar='PARAMETERNAME=VALUE',
              required=False, multiple=True,
              help='Specify a method input parameter with its value. '
                   'May be specified multiple times. '
                   'Default: No input parameters.')
@add_options(namespace_option)
@click.pass_obj
def class_invokemethod(context, classname, methodname, **options):
    """
    Invoke a method on a class.

    Invoke a static CIM method (METHODNAME argument) on a CIM class (CLASSNAME
    argument) in a CIM namespace (--namespace option), and display the method
    return value and output parameters. If no namespace was specified, the
    default namespace of the connection is used.

    The method input parameters are specified using the --parameter option,
    which may be specified multiple times.

    Pywbemcli retrieves the class definition from the server in order to
    verify that the specified input parameters are consistent with the
    parameter characteristics in the method definition.

    Use the 'instance invokemethod' command to invoke CIM methods on CIM
    instances.

    Example:

      pywbemcli -n myconn class invokemethod CIM_Foo methodx -p p1=9 -p p2=Fred
    """
    context.execute_cmd(lambda: cmd_class_invokemethod(context,
                                                       classname,
                                                       methodname,
                                                       options))


@class_group.command('enumerate', options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME', required=False)
@add_options(deep_inheritance_class_option)
@add_options(local_only_class_option)
@add_options(no_qualifiers_class_option)
@add_options(include_classorigin_class_option)
@add_options(names_only_option)
@add_options(namespace_option)
@add_options(summary_option)
@click.pass_obj
def class_enumerate(context, classname, **options):
    """
    List top classes or subclasses of a class in a namespace.

    Enumerate CIM classes starting either at the top of the class hierarchy
    in the specified CIM namespace (--namespace option), or at the specified
    class (CLASSNAME argument) in the specified namespace. If no namespace was
    specified, the default namespace of the connection is used.

    The --local-only, --include-classorigin, and --no-qualifiers options
    determine which parts are included in each retrieved class.

    The --deep-inheritance option defines whether or not the complete subclass
    hierarchy of the classes is retrieved.

    The --names-only option can be used to show only the class paths.

    In the output, the classes and class paths will be formatted as defined
    by the --output-format general option. Table formats on classes will be
    replaced with MOF format.

    Examples:

      pywbemcli -n myconn class enumerate -n interop

      pywbemcli -n myconn class enumerate CIM_Foo -n interop
    """
    context.execute_cmd(lambda: cmd_class_enumerate(context, classname,
                                                    options))


@class_group.command('references', options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME', required=True)
@click.option('--rc', '--result-class', 'result_class', type=str,
              required=False, metavar='CLASSNAME',
              help='Filter the result set by result class name. '
                   'Subclasses of the specified class also match.')
@click.option('-r', '--role', type=str, required=False,
              metavar='PROPERTYNAME',
              help='Filter the result set by source end role name.')
@add_options(no_qualifiers_class_option)
@add_options(include_classorigin_class_option)
@add_options(propertylist_option)
@add_options(names_only_option)
@add_options(namespace_option)
@add_options(summary_option)
@click.pass_obj
def class_references(context, classname, **options):
    """
    List the classes referencing a class.

    List the CIM (association) classes that reference the specified class
    (CLASSNAME argument) in the specified CIM namespace
    (--namespace option). If no namespace was specified, the default namespace
    of the connection is used.

    The classes to be retrieved can be filtered by the --role and
    --result-class options.

    The --include-classorigin, --no-qualifiers, and --propertylist options
    determine which parts are included in each retrieved class.

    The --names-only option can be used to show only the class paths.

    In the output, the classes and class paths will be formatted as defined
    by the --output-format general option. Table formats on classes will be
    replaced with MOF format.

    Examples:

      pywbemcli -n myconn class references CIM_Foo -n interop
    """
    context.execute_cmd(lambda: cmd_class_references(context, classname,
                                                     options))


@class_group.command('associators', options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME', required=True)
@click.option('--ac', '--assoc-class', 'assoc_class', type=str, required=False,
              metavar='CLASSNAME',
              help='Filter the result set by association class name. '
                   'Subclasses of the specified class also match.')
@click.option('--rc', '--result-class', 'result_class', type=str,
              required=False, metavar='CLASSNAME',
              help='Filter the result set by result class name. '
                   'Subclasses of the specified class also match.')
@click.option('-r', '--role', type=str, required=False,
              metavar='PROPERTYNAME',
              help='Filter the result set by source end role name.')
@click.option('--rr', '--result-role', 'result_role', type=str, required=False,
              metavar='PROPERTYNAME',
              help='Filter the result set by far end role name.')
@add_options(no_qualifiers_class_option)
@add_options(include_classorigin_class_option)
@add_options(propertylist_option)
@add_options(names_only_option)
@add_options(namespace_option)
@add_options(summary_option)
@click.pass_obj
def class_associators(context, classname, **options):
    """
    List the classes associated with a class.

    List the CIM classes that are associated with the specified class
    (CLASSNAME argument) in the specified CIM namespace
    (--namespace option). If no namespace was specified, the default namespace
    of the connection is used.

    The classes to be retrieved can be filtered by the --role, --result-role,
    --assoc-class, and --result-class options.

    The --include-classorigin, --no-qualifiers, and --propertylist options
    determine which parts are included in each retrieved class.

    The --names-only option can be used to show only the class paths.

    In the output, the classes and class paths will be formatted as defined
    by the --output-format general option. Table formats on classes will be
    replaced with MOF format.

    Examples:

      pywbemcli -n myconn class associators CIM_Foo -n interop
    """
    context.execute_cmd(lambda: cmd_class_associators(context, classname,
                                                      options))


@class_group.command('find', options_metavar=CMD_OPTS_TXT)
@click.argument('classname-glob', type=str, metavar='CLASSNAME-GLOB',
                required=True)
@click.option('-n', '--namespace', type=str, multiple=True,
              required=False, metavar='NAMESPACE',
              help='Add a namespace to the search scope. '
                   'May be specified multiple times. '
                   'Default: Search in all namespaces of the server.')
@add_options(multiple_namespaces_option)
@click.option('-s', '--sort', is_flag=True, required=False,
              help='Sort by namespace. Default is to sort by classname')
@click.pass_obj
def class_find(context, classname_glob, **options):
    """
    List the classes with matching class names on the server.

    Find the CIM classes whose class name matches the specified wildcard
    expression (CLASSNAME-GLOB argument) in all CIM namespaces of the
    WBEM server, or in the specified namespace (--namespace option).

    The CLASSNAME-GLOB argument is a wildcard expression that is matched on
    class names case insensitively.
    The special characters from Unix file name wildcarding are supported
    ('*' to match zero or more characters, '?' to match a single character,
    and '[]' to match character ranges). To avoid shell expansion of wildcard
    characters, the CLASSNAME-GLOB argument should be put in quotes.

    For example, "pywbem_*" returns classes whose name begins with "PyWBEM_",
    "pywbem_", etc. "*system*" returns classes whose names include the case
    insensitive string "system".

    In the output, the classes will be formatted as defined by the
    --output-format general option if it specifies table output. Otherwise
    the classes will be in the form "NAMESPACE:CLASSNAME".

    Examples:

      pywbemcli -n myconn class find "CIM_*System*" -n interop

      pywbemcli -n myconn class find *Foo*
    """
    context.execute_cmd(lambda: cmd_class_find(context, classname_glob,
                                               options))


@class_group.command('tree', options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME', required=False)
@click.option('-s', '--superclasses', is_flag=True, default=False,
              help='Show the superclass hierarchy. '
                   'Default: Show the subclass hierarchy.')
@add_options(namespace_option)
@click.pass_obj
def class_tree(context, classname, **options):
    """
    Show the subclass or superclass hierarchy for a class.

    List the subclass or superclass hierarchy of a CIM class (CLASSNAME
    argument) or CIM namespace (--namespace option):

    - If CLASSNAME is omitted, the complete class hierarchy of the specified
      namespace is retrieved.

    - If CLASSNAME is specified but not --superclasses, the class and its
      subclass hierarchy in the specified namespace are retrieved.

    - If CLASSNAME and --superclasses are specified, the class and its
      superclass ancestry up to the top-level class in the specified namespace
      are retrieved.

    If no namespace was specified, the default namespace of the connection is
    used.

    In the output, the classes will formatted as a ASCII graphical tree; the
    --output-format general option is ignored.

    Examples:

      pywbemcli -n myconn class tree -n interop

      pywbemcli -n myconn class tree CIM_Foo -n interop

      pywbemcli -n myconn class tree CIM_Foo -s -n interop
    """
    context.execute_cmd(lambda: cmd_class_tree(context, classname, options))

#####################################################################
#
#  Command functions for each of the commands in the class group
#
#####################################################################


def cmd_class_get(context, classname, options):
    """
    Get the class defined by the argument.

    Gets the class defined by CLASSNAME from the WBEM server and displays
    the class. If the class cannot be found, the server returns a CIMError
    exception.

    """
    try:
        result_class = context.conn.GetClass(
            classname,
            namespace=options['namespace'],
            LocalOnly=options['local_only'],
            IncludeQualifiers=options['no_qualifiers'],
            IncludeClassOrigin=options['include_classorigin'],
            PropertyList=resolve_propertylist(options['propertylist']))

        display_cim_objects(context, result_class,
                            output_format=context.output_format)
    except Error as er:
        raise click.ClickException("%s: %s" % (er.__class__.__name__, er))


def cmd_class_invokemethod(context, classname, methodname, options):
    """
    Create an instance and submit to a WBEM server
    """
    try:
        process_invokemethod(context, classname, methodname, options)
    except Exception as ex:
        raise click.ClickException("%s: %s" % (ex.__class__.__name__, ex))


def cmd_class_enumerate(context, classname, options):
    """
        Enumerate the classes returning a list of classes from the WBEM server.
    """
    # results may be either classes or classnames
    try:
        if options['names_only']:
            results = context.conn.EnumerateClassNames(
                ClassName=classname,
                namespace=options['namespace'],
                DeepInheritance=options['deep_inheritance'])
        else:
            results = context.conn.EnumerateClasses(
                ClassName=classname,
                namespace=options['namespace'],
                LocalOnly=options['local_only'],
                DeepInheritance=options['deep_inheritance'],
                IncludeQualifiers=options['no_qualifiers'],
                IncludeClassOrigin=options['include_classorigin'])

        display_cim_objects(context, results, context.output_format,
                            summary=options['summary'], sort=True)

    except Error as er:
        raise click.ClickException("%s: %s" % (er.__class__.__name__, er))


def cmd_class_references(context, classname, options):
    """
    Execute the references request operation to get references for
    the classname defined
    """
    if options['namespace']:
        classname = CIMClassName(classname, namespace=options['namespace'])

    try:
        if options['names_only']:
            results = context.conn.ReferenceNames(
                classname,
                ResultClass=options['result_class'],
                Role=options['role'])
        else:
            results = context.conn.References(
                classname,
                ResultClass=options['result_class'],
                Role=options['role'],
                IncludeQualifiers=options['no_qualifiers'],
                IncludeClassOrigin=options['include_classorigin'],
                PropertyList=resolve_propertylist(options['propertylist']))

        display_cim_objects(context, results, context.output_format,
                            summary=options['summary'], sort=True)

    except Error as er:
        raise click.ClickException("%s: %s" % (er.__class__.__name__, er))


def cmd_class_associators(context, classname, options):
    """
    Execute the references request operation to get references for
    the classname defined
    """
    if options['namespace']:
        classname = CIMClassName(classname, namespace=options['namespace'])

    try:
        if options['names_only']:
            results = context.conn.AssociatorNames(
                classname,
                AssocClass=options['assoc_class'],
                Role=options['role'],
                ResultClass=options['result_class'],
                ResultRole=options['result_role'])
        else:
            results = context.conn.Associators(
                classname,
                AssocClass=options['assoc_class'],
                Role=options['role'],
                ResultClass=options['result_class'],
                ResultRole=options['result_role'],
                IncludeQualifiers=options['no_qualifiers'],
                IncludeClassOrigin=options['include_classorigin'],
                PropertyList=resolve_propertylist(options['propertylist']))

        display_cim_objects(context, results, context.output_format,
                            summary=options['summary'], sort=True)

    except Error as er:
        raise click.ClickException("%s: %s" % (er.__class__.__name__, er))


def cmd_class_find(context, classname_glob, options):
    """
    Execute the command for get class and display the result. The result is
    a list of classes/namespaces
    """

    if options['namespace']:
        ns_names = options['namespace']
    else:
        try:
            ns_names = context.wbem_server.namespaces
            ns_names.sort()
        except CIMError as ce:
            # allow processing to continue if no interop namespace
            if ce.status_code == CIM_ERR_NOT_FOUND:
                click.echo('WARNING: %s' % ce)
                ns_names = [context.conn.default_namespace]

    try:
        names_dict = {}
        for ns in ns_names:
            classnames = context.conn.EnumerateClassNames(
                namespace=ns, DeepInheritance=True)
            filtered_classnames = filter_namelist(classname_glob, classnames)
            names_dict[ns] = filtered_classnames

        rows = []
        for ns_name in names_dict:
            ns_rows = [[ns_name, name] for name in names_dict[ns_name]]
            row = 0 if options['sort'] else 1
            ns_rows.sort(key=lambda x: x[row])
            rows.extend(ns_rows)

        context.spinner.stop()
        if context.output_format in TABLE_FORMATS:
            headers = ['Namespace', 'Classname']
            click.echo(format_table(rows, headers,
                                    title='Find class %s' % classname_glob,
                                    table_format=context.output_format))
        else:
            # Display function to display classnames returned with
            # their namespaces in the form <namespace>:<classname>
            context.spinner.stop()
            for row in rows:
                click.echo('  %s:%s' % (row[0], row[1]))

    except Error as er:
        raise click.ClickException("%s: %s" % (er.__class__.__name__, er))


def cmd_class_tree(context, classname, options):
    """
    Execute the command to enumerate classes from the top or starting at the
    classname argument. Then format the results to be displayed as a
    left-justified tree using the asciitree library.
    The --superclasses option determines if the superclass tree or the
    subclass tree is displayed.
    """
    try:
        if options['superclasses']:
            if classname is None:
                raise click.ClickException('CLASSNAME argument required for '
                                           '--superclasses option')

            # get the superclasses into a list
            class_ = context.conn.GetClass(classname,
                                           namespace=options['namespace'])
            classes = []
            classes.append(class_)
            while class_.superclass:
                class_ = context.conn.GetClass(class_.superclass,
                                               namespace=options['namespace'])
                classes.append(class_)
            classname = None

        else:
            # get the subclass hierarchy either complete or starting at the
            # optional CLASSNAME
            classes = context.conn.EnumerateClasses(
                ClassName=classname,
                namespace=options['namespace'],
                DeepInheritance=True)
    except Error as er:
        raise click.ClickException("%s: %s" % (er.__class__.__name__, er))

    # display the list of classes as a tree. The classname is the top
    # of the tree.
    context.spinner.stop()
    display_class_tree(classes, classname)


def cmd_class_delete(context, classname, options):
    """Delete a class from the WBEM server repository"""
    if options['namespace']:
        classname = CIMClassName(classname, namespace=options['namespace'])

    instnames = context.conn.EnumerateInstanceNames(classname)
    subclassnames = context.conn.EnumerateClassNames(ClassName=classname,
                                                     DeepInheritance=True)

    if subclassnames:
        raise click.ClickException('Delete rejected; subclasses exist')

    if not options['force']:
        if instnames:
            raise click.ClickException('Delete rejected; instances exist')
    else:
        for instname in instnames:
            context.conn.DeleteInstance(instname)

    instnames = context.conn.EnumerateInstanceNames(classname)
    if instnames:
        raise click.ClickException('Delete rejected; instance delete failed')

    try:
        context.conn.DeleteClass(classname)
        click.echo('%s delete successful' % classname)
    except Error as er:
        raise click.ClickException("%s: %s" % (er.__class__.__name__, er))



# !PROCESS!AT!STARTUP!
# The previous statement is used by pywbemcli to force this script to be
# run at pywbemcli startup and not included in the list of --mock-server
# files that are used to build the repository.  This is a development
# test aid.
# (C) Copyright 2019 IBM Corp.
# (C) Copyright 2019 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
    Python code to mock pywbemcli._common.pywbemcli_prompt. Returns the
    value defined in RETURN_VALUE

    This file was defined to mock the return from pywbemcli instance
    get/delete/... with the simple_assoc_mock_model and return "0" to represent
    the first entry in the displayed output.

    Add this file to the set of mock options and the prompt for picking
    the instance will be output but the prompt for a response from the
    user will be bypassed and the value defined in RETURN_VALUE returned.
"""
from mock import Mock

import pywbemtools
RETURN_VALUE = "3"


def mock_prompt(msg):
    """Mock function to replace pywbemcli_prompt and return a value"""
    print('MOCK_CLICK_PROMPT %s' % msg)
    return RETURN_VALUE


# pylint: disable=protected-access
pywbemtools.pywbemcli.click.prompt = Mock(side_effect=mock_prompt)



# Copyright 2018 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests the general log option
"""
import os
import pytest
from .cli_test_extensions import CLITestsBase

TEST_DIR = os.path.dirname(__file__)

SIMPLE_MOCK_FILE = 'simple_mock_model.mof'

OK = True
RUN = True
FAIL = False

TEST_CASES = [
    # desc - Description of test
    # inputs - String, or list of args or dict of 'env', 'args', 'general',
    #          and 'stdin'. See See CLITestsBase.subcmd_test()  for
    #          detailed documentation
    # exp_response - Dictionary of expected responses (stdout, stderr, rc) and
    #                test definition (test: <testname>).
    #                See CLITestsBase.subcmd_test() for detailed documentation.
    # mock - None or name of files (mof or .py),
    # condition - If True, the test is executed,  Otherwise it is skipped.

    ['Verify log of class get blah. class get that fails',
     {'general': ['-l', 'all=stderr'],
      'cmdgrp': 'class',
      'args': ['get', 'blah']},
     {'stderr': ['-pywbem.api',
                 'FakedWBEMConnection',
                 r'GetClass\(ClassName=', r'blah',
                 '-Exception:'],
      'test': 'regex',
      'rc': 1},
     SIMPLE_MOCK_FILE, OK],

    ['Verify log of class  get command get local-only. Test stdoit',
     {'general': ['-l', 'all=stderr:summary'],
      'cmdgrp': 'class',
      'args': ['get', 'CIM_Foo_sub2', '--local-only']},
     {'stdout': ['class CIM_Foo_sub2 : CIM_Foo {',
                 '',
                 '   string cimfoo_sub2;',
                 '',
                 '};', '', ],
      'stderr': ["blahblah"],
      'test': 'patterns'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify Log of class get command local-only. test stderr'
     'Cannot test stderr and stdout in same test.',
     {'general': ['-l', 'all=stderr:summary'],
      'cmdgrp': 'class',
      'args': ['get', 'CIM_Foo_sub2', '--local-only']},
     {'stderr': [r'-pywbem.api.',
                 r'FakedWBEMConnection',
                 r'-Request:',
                 r'-Return:',
                 r'GetClass\(CIMClass ', r'CIM_Foo_sub2'],
      'test': 'regex'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify log of class get command local-only. Test stderr'
     'Cannot test stderr and stdout in same test.',
     {'general': ['-l', 'api=stderr:summary'],
      'cmdgrp': 'class',
      'args': ['get', 'CIM_Foo_sub2', '--local-only']},
     {'stderr': [r'-pywbem.api.',
                 r'FakedWBEMConnection',
                 r'-Request:',
                 r'-Return:',
                 r'GetClass\(CIMClass ', r'CIM_Foo_sub2'],
      'test': 'regex'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify log http class get command local-only. Should be no log '
     'because mock does not use http'
     'Cannot test stderr and stdout in same test.',
     {'general': ['-l', 'http=stderr:summary'],
      'cmdgrp': 'class',
      'args': ['get', 'CIM_Foo_sub2', '--local-only']},
     {'stderr': [],
      'test': 'in'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify log with error in definition. Cannot test stderr and stdout in '
     'same test.',
     {'general': ['-l', 'httpx=stderr'],
      'cmdgrp': 'class',
      'args': ['get', 'CIM_Foo_sub2', '--local-only']},
     {'stderr': ["Error: Logger configuration error. input: "],
      'rc': 1,
      'test': 'regex'},
     SIMPLE_MOCK_FILE, OK],

    ['Verify log with error in definition. invalid type',
     {'general': ['-l', 'allx=stderr'],
      'cmdgrp': 'class',
      'args': ['get', 'CIM_Foo_sub2', '--local-only']},
     {'stderr': ["Error: Logger configuration error. input: allx=stderr. "
                 "Exception: Invalid simple logger name:"],
      'rc': 1,
      'test': 'regex'},
     SIMPLE_MOCK_FILE, OK],


    ['Verify invalid log parameter fails (no value) same test.',
     {'general': ['-l'],
      'cmdgrp': 'class',
      'args': ['get', 'CIM_Foo_sub2', '--local-only']},
     {'stderr': ["Usage: pywbemcli [GENERAL-OPTIONS] COMMAND [ARGS]"],
      'rc': 2,
      'test': 'in'},
     SIMPLE_MOCK_FILE, OK],

]


class TestLogOption(CLITestsBase):
    """
    Test use of the general log option. This was be tested in the
    test_general_opts.py file because it requires execution of a command
    to actually use the log and create logs.
    """

    @pytest.mark.parametrize(
        "desc, inputs, exp_response, mock, condition", TEST_CASES)
    def test_log_option(self, desc, inputs, exp_response, mock, condition):
        """
        Common test method for those subcommands and options in the
        class subcmd that can be tested.  This includes:

          * Subcommands like help that do not require access to a server

          * Subcommands that can be tested with a single execution of a
            pywbemcli command.
        """
        cmd_grp = inputs['cmdgrp'] if inputs['cmdgrp'] else ''

        self.command_test(desc, cmd_grp, inputs, exp_response,
                          mock, condition)



# Copyright 2017 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utilities to exercise pywbemcli both as a separate executable and in line with
a direct call.
"""

from __future__ import absolute_import, print_function

import os
import sys
import re
from copy import copy
from subprocess import Popen, PIPE
import six


def execute_pywbemcli(args, env=None, stdin=None, verbose=None):
    """
    Invoke the 'pywbemcli' command as a child process.

    This requires that the 'pywbemcli' command is installed in the current
    Python environment.

    Parameters:

      args (iterable of :term:`string`): Command line arguments and subcommand,
        without the command name.
        Each single argument must be its own item in the iterable; combining
        the arguments into a string does not work.
        The arguments may be binary strings, encoded in UTF-8, or unicode
        strings.

      env (dict): Environment variables to be put into the environment when
        calling the command. May be `None`. Dict key is the variable name as a
        :term:`string`; dict value is the variable value as a :term:`string`
        (without any shell escaping needed).

      stdin: (:term:`string` or None):
        Passed to the executable as stdin.

      verbose: (bool)
        If True, display args, env, and stdin before executing the command

    Returns:

      tuple(rc, stdout, stderr): Output of the command, where:

        * rc(int): Exit code of the command.
        * stdout(:term:`unicode string`): Standard output of the command,
          as a unicode string with newlines represented as '\\n'.
          Returns empty string, if there was no data.
        * stderr(:term:`unicode string`): Standard error of the command,
          as a unicode string with newlines represented as '\\n'.
          Returns empty string, if there was no data.
    """
    cli_cmd = u'pywbemcli'

    if env is None:
        env = {}
    else:
        env = copy(env)

    # TODO should we consider removing all env variables before each
    # test???

    env['PYTHONPATH'] = '.'  # Use local files
    env['PYTHONWARNINGS'] = ''  # Disable for parsing output

    # Put the env vars into the environment of the current Python process,
    # from where they will be inherited into its child processes (-> shell ->
    # cli command).
    for name in env:
        value = env[name]
        if value is None:
            if name in os.environ:
                del os.environ[name]
        else:
            os.environ[name] = value

    if verbose:
        display_envvars = '\n'.join(
            [var for var in os.environ if 'PYWBCLI' in var])
        print('OS_ENVIRON %s' % display_envvars)

    assert isinstance(args, (list, tuple))
    cmd_args = [cli_cmd]
    for arg in args:
        if not isinstance(arg, six.text_type):
            arg = arg.decode('utf-8')
        cmd_args.append(arg)

    if verbose:
        print("\nexecute_pywbemcli CMD_ARGS %s" % cmd_args)

    if stdin and six.PY3:
        stdin = stdin.encode('utf-8')

    if verbose and stdin:
        print('stdin %s' % stdin)

    # The click package on Windows writes NL at the Python level
    # as '\r\r\n' at the level of the shell under some cases. This is
    # documented in Click issue #1271 for Click 7.0 The Popen universal_newlines
    # does not handle this variation. To fix for now, remove universal_newlines
    # and replace \r\r\n with \n then \r\n and then \r also with \n in the code
    # below.
    #
    # The original code was:
    # proc = Popen(cmd_args, shell=True, stdout=PIPE, stderr=PIPE, stdin=PIPE,
    #             universal_newlines=True)
    # stout_str, stderr_str = proc.communicate(input=stdin)
    # Temp alternative is the following line with universal_newlines=False
    # and the second change to fix the EOLs ourself.
    proc = Popen(cmd_args, shell=False, stdin=PIPE, stdout=PIPE, stderr=PIPE,
                 universal_newlines=False)

    stdout_str, stderr_str = proc.communicate(input=stdin)
    rc = proc.returncode

    for name in env:
        del os.environ[name]

    if verbose:
        print('output type %s\nstdout:%r\nstderr:%r' % (type(stdout_str),
                                                        stdout_str,
                                                        stderr_str))

    if isinstance(stdout_str, six.binary_type):
        stdout_str = stdout_str.decode('utf-8')
    if isinstance(stderr_str, six.binary_type):
        stderr_str = stderr_str.decode('utf-8')

    # Replacement for Popen universal_newlines because of Click issue # 1271
    # Second part of temp patch for CRCRNL. Does what popen does and
    # CRCRNL replacement.
    if sys.platform == 'win32':
        stdout_str = stdout_str.replace('\r\r\n', '\n')  \
                               .replace('\r\n', '\n')  \
                               .replace('\r', '')
        stderr_str = stderr_str.replace('\r\r\n', '\n')  \
                               .replace('\r\n', '\n')  \
                               .replace('\r', '')

    return rc, stdout_str, stderr_str


def assert_rc(exp_rc, rc, stdout, stderr):
    """
    Assert that the specified return code is as expected.

    The actual return code is compared with the expected return code,
    and if they don't match, stdout and stderr are displayed as a means
    to help debugging the issue.

    Parameters:

      exp_rc (int): expected return code.

      rc (int): actual return code.

      stdout (string): stdout of the command, for debugging purposes.

      stderr (string): stderr of the command, for debugging purposes.
    """

    assert exp_rc == rc, \
        "Unexpected exit code (expected {}, got {})\n" \
        "  stdout:\n" \
        "{}\n\n" \
        "  stderr:\n" \
        "{}". \
        format(exp_rc, rc, stdout, stderr)


def assert_patterns(exp_patterns, act_lines, meaning):
    """
    Assert that the specified lines match the specified patterns.

    The patterns are matched against the complete line from begin to end,
    even if no begin and end markers are specified in the patterns.

    Parameters:

      exp_patterns (iterable of string): regexp patterns defining the expected
        value for each line.

      act_lines (iterable of string): the lines to be matched.

      meaning (string): A short descriptive text that identifies the meaning
        of the lines that are matched, e.g. 'stderr'.
    """
    assert len(act_lines) == len(exp_patterns), \
        "Unexpected number of lines in test desc: {}:\n" \
        "Expected line cnt={}:\n" \
        "{}\n\n" \
        "Actual line cnt={}:\n" \
        "{}\n". \
        format(meaning, len(act_lines), '\n'.join(act_lines),
               len(exp_patterns), '\n'.join(exp_patterns))

    for i, act_line in enumerate(act_lines):
        exp_line = exp_patterns[i]
        # if not exp_line.endswith('$'):
        #    exp_line += '$'
        assert re.match(exp_line, act_line), \
            "Unexpected line {} in test desc:{}:\n" \
            "Expected line vs. actual line:\n" \
            "------------\n" \
            "{}\n" \
            "------------\n" \
            "{}\n" \
            "------------\n". \
            format(i, meaning, exp_line, act_line)


def assert_lines(exp_lines, act_lines, meaning):
    """
    Assert that the specified lines match exactly the lines specified in
    exp_lines. This does not require that the pattern lines escape any
    special characters, etc.

    The exp_lines are matched against the complete line from begin to end. The
    test stops at the first difference

    Parameters:

      exp_lines (iterable of string): the expected string for each line.

      act_lines (iterable of string): the lines to be matched.

      meaning (string): A short descriptive text that identifies the meaning
        of the lines that are matched, e.g. 'stderr'.
    """
    assert len(act_lines) == len(exp_lines), \
        "Unexpected number of lines in {}:\n" \
        "Expected lines cnt={}:\n" \
        "{}\n" \
        "Actual lines cnt={}:\n" \
        "{}\n". \
        format(meaning, len(exp_lines), '\n'.join(exp_lines),
               len(act_lines), '\n'.join(act_lines))

    for i, act_line in enumerate(act_lines):
        exp_line = exp_lines[i]
        assert exp_line == act_line, \
            "Unexpected line {} in {}:\n" \
            "  expected line vs. actual line:\n" \
            "------------\n" \
            "{}\n" \
            "------------\n" \
            "{}\n" \
            "------------\n". \
            format(i, meaning, exp_line, act_line)



# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""
Version of the pywbemcli package and check for valid python versions.

The actual package version is deterrmined through pbr package through
git tag information  and is only referenced here to create the __version__
variable.
"""
import sys
import pbr.version

__all__ = ['__version__']


#: Version of the pywbemcli package, as a :term:`string`.
#:
#: Possible formats for the version string are:
#:
#: * "M.N.U.dev0": During development of future M.N.U release (not released to
#:   PyPI)
#: * "M.N.U.rcX": Release candidate X of future M.N.U release (not released to
#:   PyPI)
#: * "M.N.U": The final M.N.U release
__version__ = pbr.version.VersionInfo('pywbemtools').release_string()


# Check supported Python versions
_PYTHON_M = sys.version_info[0]
_PYTHON_N = sys.version_info[1]
if _PYTHON_M == 2 and _PYTHON_N < 7:
    raise RuntimeError('On Python 2, pywbemtools requires Python 2.7')
if _PYTHON_M == 3 and _PYTHON_N < 4:
    raise RuntimeError('On Python 3, pywbemtools requires Python 3.4 or higher')



#!/usr/bin/env python
"""
This is a simple tool to execute pywbemcli and capture the results

"""
from __future__ import absolute_import, print_function
import shlex
import os
from subprocess import Popen, PIPE


ERRORS = 0
USE_MD = False


def execute_cmd(cmd_str, shell=False):
    """
    Execute the command in cmd_str and get its exitcode, stdout and stderr.

    This function calls subprocess Popen to be able to execute the
    command and  get stdout, stderr, and the exitcode.

    Parameters:
      cmd_str (:term: `string)
        Command to be executed

      shell (boolean)
        Defines whether shell is to be used to execute the command.  Normally
        the shell should not be required.

    Return:
        Tuple of exitcode, out, err where out and err are strings representing
        the stderr and stdout data
    """
    args = shlex.split(cmd_str)

    proc = Popen(args, stdout=PIPE, stderr=PIPE, shell=shell)
    std_out_str, std_err_str = proc.communicate()
    exitcode = proc.returncode

    # return tuple of exitcode, stdout, stderr
    return exitcode, std_out_str, std_err_str


def md_headline(title, level):
    """
    Format a markdown header line based on the level argument
    """
    level_char_list = ['=', '-']
    try:
        level_char = level_char_list[level]
    except IndexError:
        level_char = '='

    return '\n%s\n%s\n' % (title, (level_char * len(title)))


def headline(title, level):
    """Output a headline that separates output groups.
    """
    if USE_MD:
        md_headline(title, level)
    else:
        print('%s\n**%s**\n' % (title, ('-' * len(title))))


def print_md_verbatum_text(text_str):
    """
    Print the text on input surrounded by the back quotes defining
    veratum text
    """
    print('```')
    print(text_str)
    print('```')


def execute_pywbemcli_cmd(cmd_str, general_options=''):
    """Output a command created from the input and test result.
    """
    global ERRORS  # pylint: disable=global-statement

    if isinstance(cmd_str, list):
        for cmd in cmd_str:
            execute_pywbemcli_cmd(cmd)
    else:
        command = 'pywbemcli %s %s' % (general_options, cmd_str)
        if USE_MD:
            print(md_headline(command, 1))
        else:
            print('%s\nPYWBEMCLI COMMAND: %s' % (('=' * 50), command))

        exitcode, out, err = execute_cmd(command)

        if USE_MD:
            print_md_verbatum_text(out)
        else:
            print(out)
        if err:
            print('**STDER:** %s' % err)

        if exitcode != 0:
            ERRORS += 1
            print('**ERROR:** cmd `%s`' % command)


#
#  The following is the input arguments for a single pywbemcli command
#

if not os.environ.get('PYWBEMCLI_SERVER'):
    print('PYWBEMCLI_SERVER env variable not found')
    quit
execute_pywbemcli_cmd("class get CIM_ManagedElement")

quit

execute_pywbemcli_cmd("class get CIM_ManagedElement -l")
execute_pywbemcli_cmd("class get CIM_ManagedElement --local-only")
execute_pywbemcli_cmd("class get CIM_ManagedElement --no-qualifiers")
execute_pywbemcli_cmd("class get CIM_ManagedElement -c")
execute_pywbemcli_cmd("class get CIM_ManagedElement --include-classorigin")
execute_pywbemcli_cmd("class get CIM_ManagedElement --include-classorigin")
execute_pywbemcli_cmd(
    "class get CIM_ManagedElement --namespace root/PG_Interop")
execute_pywbemcli_cmd("class get CIM_ManagedElement -p InstanceID -p Caption")
execute_pywbemcli_cmd("class get CIM_ManagedElement -p InstanceID")
execute_pywbemcli_cmd("class get CIM_ManagedElement -p \"\"")

execute_pywbemcli_cmd("class enumerate")
execute_pywbemcli_cmd("class enumerate CIM_System")
execute_pywbemcli_cmd("class enumerate CIM_System -d")
execute_pywbemcli_cmd("class enumerate CIM_System -l")
execute_pywbemcli_cmd("class enumerate CIM_System -s")
execute_pywbemcli_cmd("class enumerate CIM_System --local-only")
execute_pywbemcli_cmd("class enumerate CIM_System --names-only")
execute_pywbemcli_cmd("class enumerate CIM_System -o")
execute_pywbemcli_cmd("class enumerate CIM_System -o -s")

#
execute_pywbemcli_cmd("class associators CIM_System")
execute_pywbemcli_cmd("class associators CIM_System s")
execute_pywbemcli_cmd("class associators CIM_System -o")
execute_pywbemcli_cmd("class associators CIM_System -o -s")

execute_pywbemcli_cmd("class references CIM_System -o")
execute_pywbemcli_cmd("class references CIM_System -o -s")
execute_pywbemcli_cmd("class references CIM_System")
execute_pywbemcli_cmd("class references CIM_System -s")

execute_pywbemcli_cmd("class find CIM_System")
execute_pywbemcli_cmd("class find PyWBEM")

# NOTE -s means something different here
execute_pywbemcli_cmd("class hierarchy CIM_System")
execute_pywbemcli_cmd("class hierarchy CIM_System -s")
execute_pywbemcli_cmd("class hierarchy CIM_ManagedElement -s")
execute_pywbemcli_cmd("class hierarchy CIM_ManagedElement --superclasses")
execute_pywbemcli_cmd(
    "class hierarchy CIM_System --superclasses --namespace root/PG_Interop")
execute_pywbemcli_cmd("class hierarchy CIM_System --namespace root/PG_Interop")

# invoke method on a class

execute_pywbemcli_cmd("instance enumerate PyWBEM_Person")
execute_pywbemcli_cmd("instance enumerate PyWBEM_Person -p name")
execute_pywbemcli_cmd("instance enumerate PyWBEM_Person -o")

execute_pywbemcli_cmd("instance enumerate PyWBEM_Person -p name -p CreationClassName",
                      general_options='--output-format table')

# TODO: Find way to do interactive in batch. This disabled for now
# execute_pywbemcli_cmd("instance get PyWBEM_Person.?")

execute_pywbemcli_cmd(
    "instance get PyWBEM_Person.CreationClassName=PyWBEM_Person,name=Bob")

execute_pywbemcli_cmd("instance create  PyWBEM_Person -x Name=Fred -x "
                      "GivenName=Jones -x CreationClassName=PyWBEM_Person")

execute_pywbemcli_cmd("instance delete PyWBEM_Person."
                      "CreationClassName=PyWBEM_Person,Name=Fred")
# # invoke method

# # query

execute_pywbemcli_cmd("instance count")

execute_pywbemcli_cmd("server brand")
execute_pywbemcli_cmd("server info")
execute_pywbemcli_cmd("server interop")
execute_pywbemcli_cmd("server namespaces")
execute_pywbemcli_cmd("server profiles")
execute_pywbemcli_cmd("server profiles -o DMTF")
# -n broken
execute_pywbemcli_cmd("server profiles -n CPU")
